import numpy as np

'''
This module contains utility functions
'''

# =====================
# = Utility functions =
# =====================

def pdist(x, y):
    '''
    Calculate pairwise square distances between matrix x and y.
    See: https://stackoverflow.com/a/56084419/5516481
    '''
    if(x.ndim==1): x = np.array([x])

    nx, p = x.shape
    x_ext = np.empty((nx, 3*p))
    x_ext[:, :p] = 1
    x_ext[:, p:2*p] = x
    x_ext[:, 2*p:] = np.square(x)

    ny = y.shape[0]
    y_ext = np.empty((3*p, ny))
    y_ext[:p] = np.square(y).T
    y_ext[p:2*p] = -2*y.T
    y_ext[2*p:] = 1

    return x_ext.dot(y_ext)

# ==============================================
# = Activation functions and their derivatives =
# ==============================================

def sigmoid(weighted_sum):
    '''
    Returns sigmoid of the weighted sum array of a layer.
    '''
    return 1 / (1 + np.exp(-weighted_sum))

def sigmoid_prime(weighted_sum, activations):
    '''
    Returns the derivative of sigmoid w.r.t layer's weighted sum.
    '''
    return activations * (1 - activations)

def tanh(weighted_sum):
    '''
    Returns tanh of the weighted sum array of a layer.
    '''
    return np.tanh(weighted_sum)

def tanh_prime(weighted_sum, activations):
    '''
    Returns the derivative of tanh w.r.t layer's weighted sum.
    '''
    return 1 - (activations ** 2)

def leakyrelu(weighted_sum):
    '''
    Returns leaky-ReLU of the weighted sum array of a layer.
    '''
    return np.where(weighted_sum > 0, weighted_sum, 0.01 * weighted_sum)

def leakyrelu_prime(weighted_sum, activations):
    '''
    Returns the derivative of leaky-ReLU w.r.t layer's weighted sum.
    '''
    return np.where(weighted_sum > 0, 1, 0.01)

def relu(weighted_sum):
    '''
    Returns ReLU of the weighted sum array of a layer.
    '''
    return np.where(weighted_sum > 0, weighted_sum, 0)

def relu_prime(weighted_sum, activations):
    '''
    Returns the derivative of ReLU w.r.t layer's weighted sum.
    '''
    return np.where(weighted_sum > 0, 1, 0)

def softmax(weighted_sum):
    '''
    Returns softmax of the weighted sum array of a layer.
    If weighted_sum is a 2D array, then it performs softmax over each row.
    '''
    if(weighted_sum.ndim == 1):
        exps = np.exp(weighted_sum - np.max(weighted_sum))
        return exps / np.sum(exps)
    else:
        normalized = weighted_sum - np.expand_dims(np.max(weighted_sum, axis=1), axis=1)
        exps = np.exp(normalized)
        return exps / np.expand_dims(np.sum(exps, axis=1), 1)

def identity(weighted_sum):
    '''
    Returns identity of the weighted sum array of a layer.
    '''
    return weighted_sum

def identity_prime(weighted_sum, activations):
    '''
    Returns the derivative of identity w.r.t layer's weighted sum.
    '''
    return 1

def softmax_prime(weighted_sum, activations):
    '''
    Returns the derivative of softmax w.r.t layer's weighted sum.
    '''
    return activations * (1 - activations)

# ========================================
# = Cost functions and their derivatives =
# ========================================

def mse(output, target):
    '''
    Returns mean squared error cost of the output.
    '''
    return 0.5 * ((output - target) ** 2)

def mse_prime(output, target):
    '''
    Returns the derivative of the mse cost.
    '''
    return (output-target)

def cross_entropy(output, target):
    '''
    Returns cross entropy cost of the output.
    '''
    return -(target * np.log(output)) - ((1-target) * np.log(1-output))

def cross_entropy_prime(output, target):
    '''
    Returns the derivative of the cross entropy cost.
    '''
    return (output-target) / (output * (1-output))

def hinge_loss(output, target):
    '''
    Returns hinge loss of the output for SVMs.
    '''
    return np.maximum(0, 1 - target*output)

def hinge_loss_prime(output, target):
    '''
    Returns derivative of hinge loss.
    '''
    return np.where((target*output)>1, 0, -1*target)

def huber(output, target):
    '''
    Returns huber loss for dqn
    '''
    error = output - target
    
    is_small_error = np.abs(error) < 1
    
    squared_loss = np.square(error)/2
    linear_loss = np.abs(error) - 0.5
    
    return np.where(is_small_error, squared_loss, linear_loss)

def huber_prime(output, target):
    '''
    Returns derivative of huber loss.
    '''
    error = output - target
    
    is_small_error = np.abs(error) < 1

    return np.where(is_small_error, error, np.sign(error))


from abc import ABC, abstractmethod

import numpy as np

'''
Optimizers module,
REF: http://cs231n.github.io/neural-networks-3/
'''

class Optimizer(ABC):
    '''
    This class is the base class for all optimizers
    '''
    @abstractmethod
    def _optimize(self, parameter, parameter_gradient):
        pass

    @property
    @abstractmethod
    def _mlearning_rate(self):
        pass

    @_mlearning_rate.setter
    @abstractmethod
    def _mlearning_rate(self, learning_rate):
        pass

    @property
    @abstractmethod
    def _mdecay_rate(self):
        pass

    def _decay(self):
        self._mlearning_rate = self._mdecay_rate*self._mlearning_rate


class GradientDescent(Optimizer):
    '''
    This class implements gradient descent optimization.
    '''
    def __init__(self, learning_rate, decay_rate=1):
        '''
        Parameters
        ----------
            learning_rate : float
            decay_rate : float
                Decay rate for leraning rate
        '''
        self._learning_rate = learning_rate
        self._decay_rate = decay_rate

    @property
    def _mlearning_rate(self):
        return self._learning_rate

    @_mlearning_rate.setter
    def _mlearning_rate(self, learning_rate):
        self._learning_rate = learning_rate

    @property
    def _mdecay_rate(self):
        return self._decay_rate

    def _optimize(self, parameter, parameter_gradient):
        # Update and return the parameter
        return parameter - (self._learning_rate * parameter_gradient)


class Momentum(Optimizer):
    '''
    This class implements momentum optimization.
    '''
    def __init__(self, learning_rate, decay_rate=1, beta=0.9):
        '''
        Parameters
        ----------
            learning_rate : float
            decay_rate : float
                Decay rate for leraning rate
            beta : float
                Should be between 0 to 1.
        '''
        self._learning_rate = learning_rate
        self._decay_rate = decay_rate
        self._beta = beta
        self._v = 0

    @property
    def _mlearning_rate(self):
        return self._learning_rate

    @_mlearning_rate.setter
    def _mlearning_rate(self, learning_rate):
        self._learning_rate = learning_rate

    @property
    def _mdecay_rate(self):
        return self._decay_rate

    def _optimize(self, parameter, parameter_gradient):
        # Integrate v
        self._v = (self._beta*self._v) - (self._learning_rate*parameter_gradient)
        # Update and return the parameter
        return parameter + self._v


class Nesterov(Optimizer):
    '''
    This class implements neterov momentum optimization.
    '''
    def __init__(self, learning_rate, decay_rate=1, beta=0.9):
        '''
        Parameters
        ----------
            learning_rate : float
            decay_rate : float
                Decay rate for leraning rate
            beta : float
                Should be between 0 to 1.
        '''
        self._learning_rate = learning_rate
        self._decay_rate = decay_rate
        self._beta = beta
        self._v = 0
        self._v_prev = 0

    @property
    def _mlearning_rate(self):
        return self._learning_rate

    @_mlearning_rate.setter
    def _mlearning_rate(self, learning_rate):
        self._learning_rate = learning_rate

    @property
    def _mdecay_rate(self):
        return self._decay_rate

    def _optimize(self, parameter, parameter_gradient):
        # Back up before updating
        self._v_prev = self._v
        # Integrate v
        self._v = (self._beta*self._v) - (self._learning_rate*parameter_gradient)
        # Update and return the parameter
        return parameter - (self._beta*self._v_prev) + ((1+self._beta)*self._v)


class Adagrad(Optimizer):
    '''
    This class implements adagrad optmization.
    '''
    def __init__(self, learning_rate, decay_rate=1):
        '''
        Parameters
        ----------
            learning_rate : float
            decay_rate : float
                Decay rate for leraning rate
        '''
        self._learning_rate = learning_rate
        self._decay_rate = decay_rate
        self._cache = 0

    @property
    def _mlearning_rate(self):
        return self._learning_rate

    @_mlearning_rate.setter
    def _mlearning_rate(self, learning_rate):
        self._learning_rate = learning_rate

    @property
    def _mdecay_rate(self):
        return self._decay_rate

    def _optimize(self, parameter, parameter_gradient):
        # For numerical stability
        eps = 10e-8
        # Calculate cache
        self._cache += parameter_gradient**2
        # Update parameter and return
        return parameter + (-self._learning_rate*parameter_gradient)/((self._cache**0.5)+eps)


class RMSprop(Optimizer):
    '''
    This class implements RMSprop optimization.
    '''
    def __init__(self, learning_rate, decay_rate=1, beta=0.9):
        '''
        Parameters
        ----------
            learning_rate : float
            decay_rate : float
                Decay rate for leraning rate
            beta : float
                Should be between 0 to 1.
        '''
        self._learning_rate = learning_rate
        self._decay_rate = decay_rate
        self._beta = beta
        self._cache = 0

    @property
    def _mlearning_rate(self):
        return self._learning_rate

    @_mlearning_rate.setter
    def _mlearning_rate(self, learning_rate):
        self._learning_rate = learning_rate

    @property
    def _mdecay_rate(self):
        return self._decay_rate

    def _optimize(self, parameter, parameter_gradient):
        # For numerical stability
        eps = 10e-8
        # Calculate cache
        self._cache = self._beta*self._cache + (1-self._beta)*(parameter_gradient**2)
        # Update parameter and return
        return parameter + (-self._learning_rate*parameter_gradient)/((self._cache**0.5)+eps)
        
        
class Adam(Optimizer):
    '''
    This class implements adam optimization.
    '''
    def __init__(self, learning_rate, decay_rate=1, beta1=0.9, beta2=0.9):
        '''
        Parameters
        ----------
            learning_rate : float
            decay_rate : float
                Decay rate for leraning rate
            beta1 : float
                Should be between 0 to 1.
            beta2 : float
                Should be between 0 to 1.
        '''
        self._learning_rate = learning_rate
        self._decay_rate = decay_rate
        self._beta1 = beta1
        self._beta2 = beta2
        self._m = 0
        self._v = 0

    @property
    def _mlearning_rate(self):
        return self._learning_rate

    @_mlearning_rate.setter
    def _mlearning_rate(self, learning_rate):
        self._learning_rate = learning_rate

    @property
    def _mdecay_rate(self):
        return self._decay_rate

    def _optimize(self, parameter, parameter_gradient):
        # For numerical stability
        eps = 10e-8
        # Momentum
        self._m = self._beta1*self._m + (1-self._beta1)*parameter_gradient
        # RMS
        self._v = self._beta2*self._v + (1-self._beta2)*(parameter_gradient**2)
        # Update parameter
        return parameter + (-self._learning_rate*self._m)/((self._v**0.5)+eps)



import warnings
import socket

class FCEUXServer:
    '''
    Server class for making NES bots. Uses FCEUX emulator.
    Visit https://www.fceux.com for info. You will also need to
    load client lua script in the emulator.
    '''
    def __init__(self, frame_func, quit_func=None, ip='localhost', port=1234):
        '''
        Parameters
        ----------
        frame_func : function
            This function will be called every frame. The function should
            accept two argument, :code:`server` (reference to this class) 
            and :code:`frame` (number of frames executed). 
        quit_func : function
            This function will be executed when the server disconnects from
            the emulator
        ip : str
            IP address of the computer.
        port : int
            Port to listen to.
        '''
        # Eshtablish connection with client
        self._serversocket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self._serversocket.bind((ip, port))
        self._serversocket.listen(5)
        self._clientsocket, self._address = self._serversocket.accept()
        
        # This function will be called every frame
        self._on_frame_func = frame_func
        self._on_quit_func = quit_func

        self._server_info = self.recv() + ' ' + str(self._address)
        self.send('ACK')

    @property
    def info(self):
        '''
        Emulator info and lua version.
        '''
        return self._server_info

    def send(self, msg):
        '''
        Send message to lua code running on the emulator.

        Parameters
        ----------
        msg : str
        '''
        if(not type(msg) is str):
            self.quit()
            raise TypeError('Arguments have to be string')

        self._clientsocket.send(bytes(msg+'\n', 'utf-8'))

    def recv(self):
        '''
        Receive message from lua code running on emulator.

        Returns
        -------
        str
            Received message from emulator.
        '''
        return self._clientsocket.recv(4096).decode('utf-8')

    def init_frame(self):
        '''
        Signal server to prep for next frame and returns
        frame count

        Returns
        -------
        int
            Frame count
        '''
        # Receive message from client    
        frame_str = self.recv()
        if(len(frame_str) == 0): 
            self.quit('Client had quit')
        frame = int(frame_str)

        return frame

    def start(self):
        '''
        Starts the server, waits for emulator to connect.
        Calls :code:`frame_func` every frame after connection
        has been established.
        '''
        try:
            # Keep receiving messaged from FCEUX and acknowledge
            while True:
                frame = self.init_frame()
                self._on_frame_func(self, frame)
        
        except BrokenPipeError:
            self.quit('Client has quit.')
        except KeyboardInterrupt:
            self.quit()

    def frame_advance(self):
        '''
        Move to next frame, should be called at the end of 
        :code:`frame_func`.
        '''
        # Send back continue message
        self.send('CONT')

    def get_joypad(self):
        '''
        Returns
        -------
        str
            Joypad button states.
        '''
        self.send('JOYPAD')
        return self.recv()

    def set_joypad(self, up=False, down=False, left=False, 
            right=False, A=False, B=False, start=False, select=False):
        '''
        Set joypad button states.
        '''
        self.send('SETJOYPAD')
        joypad = str(up)+' '+str(down)+' '+str(left)+' '+str(right)\
            +' '+str(A)+' '+str(B)+' '+str(start)+' '+str(select)
        self.send(joypad)

    def read_mem(self, addr, signed=False):
        '''
        Read memory address.

        Parameters
        ----------
        addr : int
            The memory address to read
        signed : bool
            If :code:`True`, returns signed integer 

        Returns
        -------
        int
            The byte at the address.
        '''
        self.send('MEM')
        self.send(str(addr))
        unsigned = int(self.recv())

        if(signed):
            return unsigned-256 if unsigned>127 else unsigned
        else:
            return unsigned

    def reset(self):
        '''
        Resets the emulator, executes a power cycle.
        '''
        self.send('RES')

    def quit(self, reason=''):
        '''
        Disconnect from emulator.

        Parameters
        ----------
        reason : str
            Reason for quitting.
        '''
        if(self._on_quit_func is not None):
            self._on_quit_func()
        self._serversocket.close()
        self._clientsocket.close()
        print(reason)
        print('Server has quit.')
        exit()
    
if(__name__ == '__main__'):
    def on_frame(server, frame):
        print(frame)
        print(server.get_joypad())
        server.frame_advance()

    server = FCEUXServer(on_frame)
    print(server.info)
    server.start()




from itertools import combinations_with_replacement

import numpy as np

'''
This module contains helper functions for preprocessing data.
'''

def onehot(input_array):
    '''
    Converts input array to one-hot array.

    Parameters
    ----------
    input_array : numpy.array
        The input numpy array.
    
    Returns
    -------
    one_hot : numpy.array
        The converted onehot array.

    Example
    -------
    >>> import numpy as np
    >>> import pykitml as pk
    >>> a = np.array([0, 1, 2])
    >>> pk.onehot(a)
    array([[1., 0., 0.],
           [0., 1., 0.],
           [0., 0., 1.]])
    '''
    array = input_array.astype(int)
    one_hot = np.zeros((array.size, array.max()+1))
    one_hot[np.arange(array.size), array] = 1
    return one_hot

def onehot_cols(dataset, cols):
    '''
    Converts/replaces columns of dataset to one-hot values.

    Parameters
    ----------
    dataset : numpy.array
        The input dataset.
    cols : list
        The columns which has to be replaced/converted
        to one-hot values.

    Returns
    -------
    dataset_new : numpy.array
        The new dataset with replaced columns.

    Example
    -------
        
        >>> import pykitml as pk
        >>> import numpy as np
        >>> a = np.array([[0, 1, 2.2], [1, 2, 3.4], [0, 0, 1.1]]) 
        >>> a
        array([[0. , 1. , 2.2],
               [1. , 2. , 3.4],
               [0. , 0. , 1.1]])
        >>> pk.onehot_cols(a, cols=[0, 1])
        array([[1. , 0. , 0. , 1. , 0. , 2.2],
               [0. , 1. , 0. , 0. , 1. , 3.4],
               [1. , 0. , 1. , 0. , 0. , 1.1]])

    '''
    offset=0
    dataset_new = dataset
    for col in cols:
        onehot_colmn = onehot(dataset_new[:, col+offset])
        dataset_new = np.delete(dataset_new, col+offset, axis=1)
        dataset_new = np.insert(dataset_new, [col+offset], onehot_colmn, axis=1)
        offset += onehot_colmn.shape[1]-1 

    return dataset_new

def onehot_cols_traintest(dataset_train, dataset_test,  cols):
    '''
    Converts/replaces columns of :code:`dataset_train` and 
    :code:`dataset_test` to one-hot values.

    Parameters
    ----------
    dataset_train : numpy.array
        The training dataset.
    dataset_test : numpy.array
        The testing dataset.
    cols : list
        The columns which has to be replaced/converted
        to one-hot values.

    Returns
    -------
    dataset_train_new : numpy.array
        The new training dataset with replaced columns.
    dataset_test_new : numpy.array
        The new testing dataset with replaced columns.

    Example
    -------
        
        >>> import pykitml as pk
        >>> import numpy as np
        >>> a_train = np.array([[0, 1, 3.2], [1, 2, 3.5], [0, 0, 3.4]])
        >>> a_test = np.array([[0, 3, 3.2], [1, 2, 4.5], [1, 3, 4.5]])
        >>> a_train_onehot, a_test_onehot = pk.onehot_cols_traintest(a_train, a_test, cols=[0,1])
        >>> a_train_onehot
        array([[1. , 0. , 0. , 1. , 0. , 0. , 3.2],
               [0. , 1. , 0. , 0. , 1. , 0. , 3.5],
               [1. , 0. , 1. , 0. , 0. , 0. , 3.4]])
        >>> a_test_onehot
        array([[1. , 0. , 0. , 0. , 0. , 1. , 3.2],
               [0. , 1. , 0. , 0. , 1. , 0. , 4.5],
               [0. , 1. , 0. , 0. , 0. , 1. , 4.5]])

    '''
    # Combine the datasets
    dataset_new = np.concatenate((dataset_train, dataset_test), axis=0)
    
    # Replace columns with on hot values
    offset=0
    for col in cols:
        onehot_colmn = onehot(dataset_new[:, col+offset])
        dataset_new = np.delete(dataset_new, col+offset, axis=1)
        dataset_new = np.insert(dataset_new, [col+offset], onehot_colmn, axis=1)
        offset += onehot_colmn.shape[1]-1 

    split = dataset_train.shape[0]
    return dataset_new[:split, :], dataset_new[split:, :]

def polynomial(dataset_inputs, degree=3, cols=[]):
    '''
    Generates polynomial features from the input dataset.
    For example, if an input sample is two dimensional and of the form [a, b], 
    the degree-2 polynomial features are :code:`[a, b, a^2, ab, b^2]`, and degree-3
    polynomial features are 
    :code:`[a, b, a^2, ab, b^2, a^3, (a^2)*b, a*(b^2), b^3]`.

    Parameters
    ----------
    dataset_inputs : numpy.array
        The input dataset to generate the ploynomials from.
    degree : int
        The dgree of the polynomial.
    cols : list
        The columns to use to generate polynomial features, columns
        not in this list will be ignored. If empty (default), all columns will
        used to generate polynomial features.

    Returns
    -------
    numpy.array
        The new dataset with polynomial features.

    Example
    -------

        >>> import numpy as np
        >>> import pykitml as pk
        >>> pk.polynomial(np.array([[1, 2], [2, 3]]), degree=2)
        array([[1., 2., 1., 2., 4.],
               [2., 3., 4., 6., 9.]])
        >>> pk.polynomial(np.array([[1, 2], [2, 3]]), degree=3)
        array([[ 1.,  2.,  1.,  2.,  4.,  1.,  2.,  4.,  8.],
               [ 2.,  3.,  4.,  6.,  9.,  8., 12., 18., 27.]])
        >>> pk.polynomial(np.array([[1, 4, 5, 2], [2, 5, 6, 3]]), degree=2, cols=[0, 3])
        array([[1., 4., 5., 2., 1., 2., 4.],
               [2., 5., 6., 3., 4., 6., 9.]])
               
    '''
    # Make sure 2D array
    if(dataset_inputs.ndim == 1):
        inputs = np.array([dataset_inputs])
    else:
        inputs = dataset_inputs

    # Choose the columns to genrate polynomial features for
    if(len(cols) == 0): cols = range(inputs.shape[1])
    
    poly_dataset = inputs

    # Generate degree terms
    for d in range(2, degree+1):
        # Generate terms indices for degree d
        term_indices = list(combinations_with_replacement(cols, r=d))
        # Multiply them to form the term and concatenate
        for indices in term_indices:
            term = inputs[:, indices].prod(axis=1)
            temp = np.zeros((poly_dataset.shape[0], poly_dataset.shape[1]+1))
            temp[:, :-1] = poly_dataset
            temp[:, -1] = term
            poly_dataset = temp

    return poly_dataset.squeeze() 




from ._functions import pdist
import numpy as np

def smote(minority_data_points, k=1):
    '''
    SMOTE (Synthetic Minority Oversampling TEchnique).
    Used to generate more data points for minority class or imbalanced learning.

    Parameters
    ----------
    minority_data_points : numpy.array
        Inputs or data points corresponding to the minority class.
    k : int
        Number of neighbors to consider.

    Returns
    -------
    new_points : numpy.array
        New generated data points (Excluding data points passed to the 
        function). :code:`k*minority_data_points.shape[0]` points will be
        generated.
    '''
    npoints = minority_data_points.shape[0]
    nfeatures = minority_data_points.shape[1]
    
    # Calculate distance between each point and evry other point
    distances = pdist(minority_data_points, minority_data_points)
    
    # Get indices of closest k neigbours for each point
    indices = np.argsort(distances, axis=1)[:, 1:k+1]

    # Get the closest k neighbours for each point
    neighbours = minority_data_points[indices].squeeze()
    neighbours = neighbours.reshape(k*npoints, nfeatures)

    # Calculate diffrence between points and k neighbours
    minority_data_points_dups = minority_data_points[np.tile(np.arange(npoints).reshape(npoints, 1), k)]
    minority_data_points_dups = minority_data_points_dups.reshape(k*npoints, nfeatures)
    diff = neighbours - minority_data_points_dups

    # Create new data points
    random_floats = np.random.uniform(0, 1, (npoints*k))
    new_points = minority_data_points_dups + (diff.T*random_floats).T

    return new_points



import os
from urllib import request

import numpy as np

from .. import pklhandler

'''
This module contains helper functions to download the adult dataset.
'''

def get():
    '''
    Downloads adult dataset from
    https://archive.ics.uci.edu/ml/datasets/adult
    and saves it as a pkl files `adult.data.pkl` and `adult.test.pkl`.

    Raises
    ------
        urllib.error.URLError
            If internet connection is not available or the URL is not accessible.
        OSError
            If the files cannot be created due to a system-related error.
        KeyError
            If invalid/unknown type.

    Note
    ----
    You only need to call this method once, i.e, after the dataset has been downloaded
    and you have the `adult.data.pkl` and `adult.test.pkl` files, you don't need to call 
    this method again.
    '''
    # Dictionary to store categorical values
    workclass_dict = {
        'Private':0, 'Self-emp-not-inc':1, 'Self-emp-inc':2, 'Federal-gov':3, 
        'Local-gov':4, 'State-gov':5, 'Without-pay':6, 'Never-worked':7
    }

    education_dict = {
        'Bachelors':0, 'Some-college':1, '11th':2, 'HS-grad':3, 'Prof-school':4, 'Assoc-acdm':5,
        'Assoc-voc':6, '9th':7, '7th-8th':8, '12th':9, 'Masters':10, '1st-4th':11, '10th':12,
        'Doctorate':13, '5th-6th':14, 'Preschool':15
    }

    marital_dict = {
        'Married-civ-spouse':0, 'Divorced':1, 'Never-married':2, 'Separated':3, 
        'Widowed':4, 'Married-spouse-absent':5, 'Married-AF-spouse':6
    }

    occupation_dict = {
        'Tech-support':0, 'Craft-repair':1, 'Other-service':2, 'Sales':3, 'Exec-managerial':4, 
        'Prof-specialty':5, 'Handlers-cleaners':6, 'Machine-op-inspct':7, 'Adm-clerical':8, 
        'Farming-fishing':9, 'Transport-moving':10, 'Priv-house-serv':11, 'Protective-serv':12, 
        'Armed-Forces':13
    }

    relationship_dict = {
        'Wife':0, 'Own-child':1, 'Husband':2, 'Not-in-family':3, 'Other-relative':4, 
        'Unmarried':5
    }

    race_dict = {
        'White':0, 'Asian-Pac-Islander':1, 'Amer-Indian-Eskimo':2, 'Other':3, 'Black':4
    }

    sex_dict = {
        'Female':0, 'Male':1
    }

    country_dict = {
        'United-States':0, 'Cambodia':1, 'England':2, 'Puerto-Rico':3, 'Canada':4, 'Germany':5, 
        'Outlying-US(Guam-USVI-etc)':6, 'India':7, 'Japan':8, 'Greece':9, 'South':10, 'China':11, 'Cuba':12, 
        'Iran':13, 'Honduras':14, 'Philippines':15, 'Italy':16, 'Poland':17, 'Jamaica':18, 'Vietnam':19, 'Mexico':20, 
        'Portugal':21, 'Ireland':22, 'France':23, 'Dominican-Republic':24, 'Laos':25, 'Ecuador':26, 'Taiwan':27, 
        'Haiti':28, 'Columbia':29, 'Hungary':30, 'Guatemala':31, 'Nicaragua':32, 'Scotland':33, 'Thailand':34, 
        'Yugoslavia':35, 'El-Salvador':36, 'Trinadad&Tobago':37, 'Peru':38, 'Hong':39, 'Holand-Netherlands':40,
    }

    out_dict = {
        '<=50K\n':0, '>50K\n':1
    }

    def download(files_name):
        # Url to download the dataset from
        url = 'https://archive.ics.uci.edu/ml/machine-learning-databases/adult/'+files_name

        # Download the dataset
        print('Downloading ' + files_name + '...')
        request.urlretrieve(url, files_name)
        print('Download complete.')

        # Parse data and save it as a pkl files
        data_array = []
        # Open the files and put the values in a list
        with open(files_name, 'r') as datafiles:
            for line in datafiles:
                try:
                    values = line.replace(' ', '').replace('.', '').split(',')
                    del values[4]
                    values[0] = float(values[0])
                    values[1] = workclass_dict[values[1]]
                    values[2] = float(values[2])
                    values[3] = education_dict[values[3]]
                    values[4] = marital_dict[values[4]]
                    values[5] = occupation_dict[values[5]]
                    values[6] = relationship_dict[values[6]]
                    values[7] = race_dict[values[7]]
                    values[8] = sex_dict[values[8]]
                    values[9] = float(values[9])
                    values[10] = float(values[10])
                    values[11] = float(values[11])
                    values[12] = country_dict[values[12]]
                    values[13] = out_dict[values[13]]
                    data_array.append(values)
                except (KeyError, ValueError, IndexError):
                    continue
        # Convert to numpy array
        np_data = np.array(data_array)

        # save it as a pkl files
        pklhandler.save(np_data, files_name+'.pkl')

    # Download the data set
    download('adult.data')
    download('adult.test')

    # Clean up
    print('Deleting unnecessary files...')
    os.remove('adult.data')
    os.remove('adult.test')


def load():
    '''
    Loads the adult dataset from `adult.data.pkl` and `adult.test.pkl` files.
    The inputs have the following columns:

    - age
    - workclass : 
      Private=0, Self-emp-not-inc=1, Self-emp-inc=2, Federal-gov=3, 
      Local-gov=4, State-gov=5, Without-pay=6, Never-worked=7
    - fnlwgt 
    - education :
      Bachelors=0, Some-college=1, 11th=2, HS-grad=3, Prof-school=4, Assoc-acdm=5,
      Assoc-voc=6, 9th=7, 7th-8th=8, 12th=9, Masters=10, 1st-4th=11, 10th=12,
      Doctorate=13, 5th-6th=14, Preschool=15
    - marital-status :
      Married-civ-spouse=0, Divorced=1, Never-married=2, Separated=3, 
      Widowed=4, Married-spouse-absent=5, Married-AF-spouse=6
    - occupation :
      Tech-support=0, Craft-repair=1, Other-service=2, Sales=3, Exec-managerial=4, 
      Prof-specialty=5, Handlers-cleaners=6, Machine-op-inspct=7, Adm-clerical=8, 
      Farming-fishing=9, Transport-moving=10, Priv-house-serv=11, Protective-serv=12, 
      Armed-Forces=13
    - relationship :
      Wife=0, Own-child=1, Husband=2, Not-in-family=3, Other-relative=4, 
      Unmarried=5
    - race :
      White=0, Asian-Pac-Islander=1, Amer-Indian-Eskimo=2, Other=3, Black=4
    - sex :
      Female=0, Male=1
    - capital-gain
    - capital-loss
    - hours-per-week
    - native-country
      United-States=0, Cambodia=1, England=2, Puerto-Rico=3, Canada=4, Germany=5, 
      Outlying-US(Guam-USVI-etc)=6, India=7, Japan=8, Greece=9, South=10, China=11, Cuba=12, 
      Iran=13, Honduras=14, Philippines=15, Italy=16, Poland=17, Jamaica=18, Vietnam=19, Mexico=20, 
      Portugal=21, Ireland=22, France=23, Dominican-Republic=24, Laos=25, Ecuador=26, Taiwan=27, 
      Haiti=28, Columbia=29, Hungary=30, Guatemala=31, Nicaragua=32, Scotland=33, Thailand=34, 
      Yugoslavia=35, El-Salvador=36, Trinadad&Tobago=37, Peru=38, Hong=39, Holand-Netherlands=40,

    The outputs are:

    - <=50K = 0/False 
    - >50K = 1/True

    Returns
    -------
    inputs_train : numpy.array
        392106x13 numpy array containing training inputs.
    outputs_train : numpy.array
        Numpy array of size 392106.
    inputs_test : numpy.array
        195780x13 numpy array containing testing inputs.
    outputs_test : numpy.array
        Numpy array of size 195780.

    Raises
    ------
        filesNotFoundError
            If `adult.data.pkl` or `adult.test.pkl` files does not exist, 
            i.e, if the dataset was not downloaded and saved using the 
            :py:func:`~get` method.  
    '''
    train = pklhandler.load('adult.data.pkl')
    inputs_train = train[:, :-1]
    outputs_train = train[:, -1]

    test = pklhandler.load('adult.test.pkl')
    inputs_test = test[:, :-1]
    outputs_test = test[:, -1]

    return inputs_train, outputs_train, inputs_test, outputs_test




import numpy, ctypes, multiprocessing

'''
This module contains helper functions to share
numpy arrays between python multiprocessing processes.

See: https://stackoverflow.com/a/5034106/5516481
'''

_ctypes_to_numpy = {
    ctypes.c_char   : numpy.dtype(numpy.uint8),
    ctypes.c_wchar  : numpy.dtype(numpy.int16),
    ctypes.c_byte   : numpy.dtype(numpy.int8),
    ctypes.c_ubyte  : numpy.dtype(numpy.uint8),
    ctypes.c_short  : numpy.dtype(numpy.int16),
    ctypes.c_ushort : numpy.dtype(numpy.uint16),
    ctypes.c_int    : numpy.dtype(numpy.int32),
    ctypes.c_uint   : numpy.dtype(numpy.uint32),
    ctypes.c_long   : numpy.dtype(numpy.int64),
    ctypes.c_ulong  : numpy.dtype(numpy.uint64),
    ctypes.c_float  : numpy.dtype(numpy.float32),
    ctypes.c_double : numpy.dtype(numpy.float64)
}

_numpy_to_ctypes = dict(zip(_ctypes_to_numpy.values(), _ctypes_to_numpy.keys()))


def shm_as_ndarray(mp_array, shape = None):
    '''
    Given a multiprocessing.Array, returns an ndarray pointing to
    the same data.
    '''

    # support SynchronizedArray:
    if not hasattr(mp_array, '_type_'):
        mp_array = mp_array.get_obj()

    dtype = _ctypes_to_numpy[mp_array._type_]
    result = numpy.frombuffer(mp_array, dtype)

    if(shape is not None):
        result = result.reshape(shape)

    return numpy.asarray(result)


def ndarray_to_shm(array, lock = False):
    '''
    Generate an 1D multiprocessing.Array containing the data from
    the passed ndarray. The data will be *copied* into shared
    memory.
    '''

    array1d = array.ravel(order='A')

    try:
        c_type = _numpy_to_ctypes[array1d.dtype]
    except KeyError:
        c_type = _numpy_to_ctypes[numpy.dtype(array1d.dtype)]

    result = multiprocessing.Array(c_type, array1d.size, lock=lock)
    shm_as_ndarray(result)[:] = array1d
    return result


import numpy as np

class PCA:
    '''
    This class implements Principle Component Analysis.
    '''
    def __init__(self, data_points, no_components):
        '''
        This class implements Principle Component Analysis, used for 
        dimensionality reduction.

        Parameters
        ----------
        data_points : numpy.array
            The dataset to perform PCA i.e. dimensionality reduction on.
        no_components : int
            Number of principle components to use.
        '''
        # Calculate covariance matrix
        covariance_matrix = (data_points.T) @ data_points;
        covariance_matrix = covariance_matrix/data_points.shape[0]

        # Perform Singular Value Decomposition on the comvariance matrix
        u, s, v = np.linalg.svd(covariance_matrix, full_matrices=True)

        # Calculate amount of variance retained
        self._retention = np.sum(s[0:no_components])/np.sum(s)

        # The transformation matrix for PCA
        self._transform = u[:, 0:no_components]

    def transform(self, data_points):
        '''
        Transforms the input dataset to lower dimensions.
        
        Parameters
        ----------
        data_points : numpy.array
            The input dataset.

        Returns
        -------
        transformed_data_points : numpy.array
            The transformed input.
        '''
        # Transform the datapoints using principle components
        return data_points@self._transform

    def inverse_transform(self, pca_points):
        '''
        Gets the original dataset from transformed points.

        Parameters
        ----------
        pca_points : numpy.array
            The trasformed points.

        '''
        # Transform from principle components back to approx feature
        return pca_points @ (self._transform.T)

    @property
    def retention(self):
        '''
        Returns the amount of variance retained, between 0 and 1.
        '''
        return round(self._retention, 2)


class InvalidFeatureType(Exception):
    '''
    Raised when specified feature type is invalid for the model.
    '''
    pass

class InvalidDistributionType(Exception):
    '''
    Raised when specified distribution type is invalid for the model.
    '''
    pass

def _valid_list(input_list, valid_items):
    '''
    Used to check if items in a list are valid.

    Parameters
    ----------
    input_list : list
        The list to check/validate.
    valid_items : list
        List of valid items the list can contain.
    '''
    if (all(item in valid_items for item in input_list) and len(input_list)>0):
        return True
    else:
        return False
        


from abc import ABC, abstractmethod
import itertools

import numpy as np
import matplotlib.pyplot as plt

import tqdm

from . import _heatmap
from . import preprocessing

class Classifier(ABC):
    '''
    Mix-in class for classifier models.
    '''

    @abstractmethod
    def get_output(self):
        '''
        Returns the output activations of the model.

        Returns
        -------
        numpy.array
            The output activations.
        '''
        pass

    @abstractmethod
    def feed(self, input_data):
        '''
        Accepts input array and feeds it to the model.

        Parameters
        ----------
        input_data : numpy.array
            The input to feed the model.

        Raises
        ------
        ValueError
            If the input data has invalid dimensions/shape.      

        Note
        ----
        This function only feeds the input data, to get the output after calling this
        function use :py:func:`get_output` or :py:func:`get_output_onehot`
        '''
        pass

    @property
    @abstractmethod
    def _out_size(self):
        '''
        Returns number of nodes/neurons in the output layer.
        '''
        pass

    def get_output_onehot(self):
        '''
        Returns the output layer activations of the model as a one-hot array. A one-hot array
        is an array of bits in which only `one` of the bits is high/true. In this case, the
        corresponding bit to the neuron/node having the highest activation will be high/true.
        
        Returns
        -------
        numpy.array
            The one-hot output activations array. 
        '''
        output_targets = self.get_output()

        if(self._out_size == 1):
            # For binary classification
            return np.where(output_targets>0.5, 1, 0)
        elif(output_targets.ndim == 1):
            # If output is a vector/1D array, axis=1 will not work
            index = np.argmax(output_targets)
            output_onehot = np.zeros((self._out_size))
            output_onehot[index] = 1
        else:
            # Create a onehot array from outputs
            output_onehot = np.zeros(output_targets.shape)
            output_onehot[np.arange(output_targets.shape[0]), np.argmax(output_targets, axis=1)] = 1

        # Return one hot array
        return output_onehot

    def accuracy(self, testing_data, testing_targets):
        '''
        Tests the accuracy of the model on the testing data passed to the
        function. This function should be only used for classification.

        Parameters
        ----------
        testing_data : numpy.array
            numpy array containing testing data.
        testing_targets : numpy.array
            numpy array containing testing targets, corresponding to the testing data.
        
        Returns
        -------
        accuracy : float
           The accuracy of the model over the testing data i.e how many testing examples
           did the model predict correctly.
        '''
        self._on_test_start()

        # Evalute over all the testing data and get outputs
        self.feed(testing_data)

        # Create a onehot array from outputs
        output_onehot = self.get_output_onehot()

        # Calculate how many examples it classified correctly
        if(self._out_size == 1):
            no_correct = (testing_targets == output_onehot).sum()
        else:    
            no_correct = (testing_targets == output_onehot).all(axis=1).sum()
        
        # Calculate accuracy
        accuracy = (no_correct/testing_data.shape[0]) * 100

        self._on_test_end()
        # return accuracy
        return round(accuracy, 2) 

    def confusion_matrix(self, test_data, test_targets, gnames=[], plot=True,):   
        '''
        Returns and plots confusion matrix on the given test data.

        Parameters
        ----------
        test_data : numpy.array
            Numpy array containing test data
        test_targets : numpy.array
            Numpy array containing the targets corresponding to the test data.
        plot : bool
            If set to false, will not plot the matrix. Default is true.
        gnames : list
            List of string names for each class/group.

        Returns
        -------
        confusion_matrix : numpy.array
            The confusion matrix. 
        '''
        print('Creating Confusion Matrix...')
        self._on_test_start()

        # feed the data
        self.feed(test_data)

        # Get output
        outputs = self.get_output_onehot()

        # Binary classification
        if(self._out_size == 1):
            # Number of groups
            ngroups = 2
            # Column/Row labels
            labels = ['False', 'True']
            # Split outputs to two groups
            outputs = preprocessing.onehot(outputs)
            targets = preprocessing.onehot(test_targets)
            # Prevent bugs that show up when outputs are all zeros
            if(outputs.shape[1] == 1):
                outputs = np.pad(outputs, ((0, 0), (0 ,1)), 'constant', constant_values=0)

        # Multiclass classification
        else:
            # Number of groups
            ngroups = self._out_size
            # Column/Row labels
            if(len(gnames) != 0):
                labels = gnames
            else:
                labels = [str(x) for x in range(0, ngroups)]
            # Targets
            targets = test_targets
        
        # Confusion matrix
        conf_mat = np.zeros((ngroups, ngroups))

        # Loop through every possibility and count them
        pairs = list(itertools.product(range(0, ngroups), repeat=2))
        for predicted, actual in tqdm.tqdm(pairs, ncols=80, unit='pairs'):
            # Make one hot vector for predicted
            pred_vec = np.zeros((ngroups))
            pred_vec[predicted] = 1

            # Make one hot vector for actual
            act_vec = np.zeros((ngroups))
            act_vec[actual] = 1

            # Count
            out_count = np.all(outputs == pred_vec, axis=1)
            target_count  = np.all(targets == act_vec, axis=1)
            tot_count = np.logical_and(out_count, target_count).sum()
            conf_mat[predicted][actual] = tot_count

        # Plot the confusion matrix
        if(plot):
            # Plot confusion matrix
            fig, ax = plt.subplots(figsize=(10, 7))
            im, _ = _heatmap.heatmap(conf_mat, labels, labels, ax=ax,
                            cmap="YlGn", cbarlabel="Count")
            _heatmap.annotate_heatmap(im, valfmt="{x:.0f}")

            # Labels
            fig.canvas.set_window_title('Confusion Matrix') 
            ax.set_xlabel('Actual')
            ax.xaxis.set_label_position('top') 
            ax.set_ylabel('Predicted')

            # Show
            plt.show()

        # return
        self._on_test_end()
        return conf_mat

    def _on_test_start(self):
        '''
        This method will be called before testing, 
        override this method needed.
        '''
        pass

    def _on_test_end(self):
        '''
        This method will be called after testing, 
        override this method needed.
        '''
        pass


import os
from urllib import request

import numpy as np
from numpy import genfromtxt

from .. import pklhandler

'''
This module contains helper functions to download and load
the boston housing dataset.
'''

def get():
    '''
    Downloads the boston dataset from
    https://archive.ics.uci.edu/ml/machine-learning-databases/housing/
    and saves it as a pkl file `boston.pkl`.
    
    Raises
    ------
        urllib.error.URLError
            If internet connection is not available or the URL is not accessible.
        OSError
            If the file cannot be created due to a system-related error.
        KeyError
            If invalid/unknown type.

    Note
    ----
    You only need to call this method once, i.e, after the dataset has been downloaded
    and you have the `boston.pkl` file, you don't need to call this method again.   
    '''
    # Url to download the dataset from
    url = 'https://archive.ics.uci.edu/ml/machine-learning-databases/housing/housing.data'
    
    # Download the dataset
    print('Downloading housing.data...')
    request.urlretrieve(url, 'housing.data')
    print('Download complete.')

    # Parse the data and save it as a pkl file
    pklhandler.save(genfromtxt('housing.data'), 'boston.pkl')
    # Delete unnecessary files
    os.remove('housing.data')
    print('Deleted unnecessary files.')

def load():
    '''
    Loads the boston housing dataset from pkl file.

    The inputs have following columns:

    - CRIM :
      per capita crime rate by town
    - ZN :
      proportion of residential land zoned for lots over 25,000 sq.ft.
    - INDUS :
      proportion of non-retail business acres per town
    - CHAS :    
      Charles River dummy variable (= 1 if tract bounds river; 0 otherwise)
    - NOX :      
      nitric oxides concentration (parts per 10 million)
    - RM :       
      average number of rooms per dwelling
    - AGE :      
      proportion of owner-occupied units built prior to 1940
    - DIS :      
      weighted distances to five Boston employment centres
    - RAD :      
      index of accessibility to radial highways
    - TAX :     
      full-value property-tax rate per $10,000
    - PTRATIO : 
      pupil-teacher ratio by town
    - B :       
      1000(Bk - 0.63)^2 where Bk is the proportion of black by town
    - LSTAT :   
      % lower status of the population
    
    The outputs are

    - MEDV :
      Median value of owner-occupied homes in $1000's

    Returns
    -------
    inputs_train : numpy.array
    outputs_train : numpy.array
    inputs_test : numpy.array
    outputs_test : numpy.array

    '''
    data_array = pklhandler.load('boston.pkl')

    inputs_train = data_array[0:500, :-1]
    outputs_train = data_array[0:500, -1]
    inputs_test = data_array[500:, :-1]
    outputs_test = data_array[500:, -1]

    return inputs_train, outputs_train, inputs_test, outputs_test


import numpy as np

from ._single_layer_model import SingleLayerModel
from ._regressor import Regressor
from . import _functions

class LinearRegression(SingleLayerModel, Regressor):
    '''
    Implements linear regression.
    '''

    @property
    def _activ_func(self):
        return _functions.identity

    @property
    def _activ_func_prime(self):
        return _functions.identity_prime

    @property
    def _cost_func(self):
        return _functions.mse

    @property
    def _cost_func_prime(self):
        return _functions.mse_prime


from abc import ABC, abstractmethod

import numpy as np

from ._minimize_model import MinimizeModel

class SingleLayerModel(MinimizeModel, ABC):
    '''
    General base class for single layer models.
    '''

    def __init__(self, input_size, output_size, reg_param=0):
        '''
        Parameters
        ----------
        input_size : int
            Size of input data or number of input features.
        output_size: int
            Number of categories or groups.
        reg_param : int
            Regularization parameter for the model, also known as 'weight decay'.
        '''        
        # Save sizes
        self._input_size = input_size
        self._output_size = output_size

        # Initialize regularization parameter
        self._reg_param = reg_param
        self._reg_param_half = reg_param/2
    
        # Initialize weights and parameters
        epsilon = np.sqrt(6)/(np.sqrt(output_size) + np.sqrt(input_size))
        weights = np.random.rand(output_size, input_size)*2*epsilon - epsilon
        biases = np.random.rand(output_size) * 2 * epsilon - epsilon
        
        # Numpy array to store activations
        self._inputa = np.array([])
        self.a = np.array([])
        self.z = np.array([])

        # Put parameters in numpy dtype=object array
        W = 0 # Weights
        B = 1 # Biases
        self._params = np.array([None, None], dtype=object)
        self._params[W] = weights
        self._params[B] = biases

    @property
    def _mparams(self):
        return self._params
    
    @_mparams.setter
    def _mparams(self, mparams):
        self._params = mparams

    @property
    def _cost_function(self):
        return self._cost_func

    @property
    def _out_size(self):
        return self._output_size

    def feed(self, input_data):
        # Constants
        W = 0 # Weights
        B = 1 # Biases

        # feed
        self._inputa = input_data
        self.z = (input_data @ self._params[W].T) + self._params[B]
        self.a = self._activ_func(self.z)

    def get_output(self):
        return self.a.squeeze()

    def _backpropagate(self, index, target):
        # Constants
        W = 0 # Weights
        B = 1 # Biases

        # Gradients
        da_dz = self._activ_func_prime(self.z[index], self.a[index])
        dc_db = self._cost_func_prime(self.a[index], target) * da_dz
        dc_dw = np.multiply.outer(dc_db, self._inputa[index])
        
        # Add regularization
        dc_dw += self._reg_param*self._params[W]
        
        # Return gradient
        gradient = np.array([None, None], dtype=object)
        gradient[W] = dc_dw
        gradient[B] = dc_db
        return gradient

    @property
    def bptt(self):
        return False

    def _get_norm_weights(self):
        W = 0
        return self._reg_param_half*(self._params[W]**2).sum()

    @property
    @abstractmethod
    def _activ_func(self):
        pass

    @property
    @abstractmethod
    def _activ_func_prime(self):
        pass

    @property
    @abstractmethod
    def _cost_func(self):
        pass

    @property
    @abstractmethod
    def _cost_func_prime(self):
        pass


import os
from urllib import request

import numpy as np

from .. import pklhandler

'''
This module contains helper functions to download and load
the S1 clustering dataset. 
'''

def get():
    '''
    Downloads the s1 clustering dataset from 
    http://cs.joensuu.fi/sipu/datasets/
    and save is as a pkl file `s1.pkl`.

    Raises
    ------
        urllib.error.URLError
            If internet connection is not available or the URL is not accessible.
        OSError
            If the file cannot be created due to a system-related error.
        KeyError
            If invalid/unknown type.

    Note
    ----
    You only need to call this method once, i.e, after the dataset has been downloaded
    and you have the `s1.pkl` file, you don't need to call this method again.
    '''
    # Url to download the dataset from
    url = 'http://cs.joensuu.fi/sipu/datasets/s1.txt'

    # Download the dataset
    print('Downloading s1.txt...')
    request.urlretrieve(url, 's1.txt')
    print('Download complete.')

    # Parse the data and save it as a pkl file
    data_array = np.loadtxt('s1.txt')
    pklhandler.save(data_array, 's1.pkl')

    # Delete unnecessary files.
    os.remove('s1.txt')
    print('Deleted unnecessary files.')

def load():
    '''
    Loads x, y points from the s1 clustering dataset from saved pickle file `s1.pkl` to 
    numpy array. S1 clustering dataset contains 15 clusters.

    Returns
    -------
    training_data : numpy.array
        5000x2 numpy array containing x, y points. 

    Raises
    ------
        FileNotFoundError
            If `s1.pkl` file does not exist, i.e, if the dataset was not 
            downloaded and saved using the :py:func:`~get` method.  
    '''
    return pklhandler.load('s1.pkl')



import numpy as np

from ._single_layer_model import SingleLayerModel
from ._classifier import Classifier
from . import _functions

class LogisticRegression(SingleLayerModel, Classifier):
    '''
    Implements logistic regression for classification.
    '''
    def __init__(self, input_size, output_size, reg_param=0):
        # Initialize base class
        super(LogisticRegression, self).__init__(input_size, output_size, reg_param)

        # Choose output activation function
        if(output_size == 1):
            # For binary classification
            self._afunc = _functions.sigmoid
            self._afunc_prime = _functions.sigmoid_prime
        else:
            # For multiclass classification
            self._afunc = _functions.softmax
            self._afunc_prime = _functions.softmax_prime

    @property
    def _activ_func(self):
        return self._afunc

    @property
    def _activ_func_prime(self):
        return self._afunc_prime

    @property
    def _cost_func(self):
        return _functions.cross_entropy

    @property
    def _cost_func_prime(self):
        return _functions.cross_entropy_prime


import numpy as np

from ._classifier import Classifier
from ._regressor import Regressor
from . import _functions

class NearestNeighbor(Classifier, Regressor):
    '''
    This class implements nearest neighbor classifier.
    '''

    def __init__(self, inputs_size, output_size, no_neighbors=1):
        '''
        Parameters
        ----------
        input_size : int
            Size of input data or number of input features.
        output_size : int
            Number of categories or groups.
        no_neighbors : int
            The number of nearest neighbors to consider.
        '''
        self._k = no_neighbors
        self._output = None

        self._input_size = inputs_size
        self._output_size = output_size

    @property
    def _out_size(self):
        return self._output_size

    def train(self, training_data, targets):
        '''
        Trains the model on the training data.

        Parameters
        ----------
        training_data : numpy.array
            numpy array containing training data.
        targets : numpy.array
            numpy array containing training targets, corresponding to the training data.
        '''
        self._inputs = training_data
        self._outputs = targets

    def feed(self, input_data):
        # Make sure array is 2D
        if(input_data.ndim == 1): input_data = np.array([input_data])

        # Get pair wise distances
        distances = _functions.pdist(input_data, self._inputs)

        # Sort the distances
        indices = np.argsort(distances, axis=1)[:, 0:self._k]

        # Get output
        self._output = np.mean(self._outputs[indices], axis=1)

    def get_output(self):
        return self._output.squeeze()
    



import random
import math


class RandomSearch:
    '''
    This class is used to search for hyperparameters. 
    '''

    def __init__(self):
        self._curr_cost = None
        self._best = False

    @property
    def best(self):
        '''
        If the last generated hyperparameters is the best so far.

        Note
        ----
        This property has to be used AFTER calling :py:func:`set_cost`
        '''
        return self._best

    def set_cost(self, cost):
        '''
        Set the cost for current hyperparameter.

        Parameters
        ----------
        cost : float
            The cost corresponding to current set of hyperparameters.
        '''
        self._curr_cost = cost

    def search(self, nsamples, nzoom, zoomratio, *args):
        '''
        Generator function to loop through randomly generated hyperparameters.
        Total number of hyperparameters sampled will be :code:`nsamples*nzoom`.
        First :code:`nsamples` points will be sampled, then the function will 
        'zoom in' around the best sample, and :code:`nsamples` more points will 
        be sampled. This will be repeated :code:`nzoom` times.
        The range for each hyperparameter should be passed as a list to
        :code:`*args`. The range should be :code:`[from, to, 'type']`, 
        for e.g. :code:`[0.8, 1, 'float']`. Three range types are available, 
        :code:`'float'`, :code:`'int'`, :code:`'log'`.

        Parameters
        ----------
        nsamples : int
            Number of hyperparameters to sample.
        nzoom : int
            Number of times to zoom in.
        zoomratio : float
            How much to zoom in.
        *args
            Range type for each hyperparameter.
        '''
        best_params = None
        min_cost = float('inf')
        range_types = args

        for z in range(nzoom):
            for i in range(nsamples):
                params = []
                # Generate hyperparameters
                for rtype in range_types:
                    l = rtype[0]
                    u = rtype[1]
                    if(rtype[2] == 'int'): params.append(random.randint(int(l), int(u)))
                    elif(rtype[2] == 'float'): params.append(random.uniform(l, u))
                    elif(rtype[2] == 'log'): params.append(10**random.uniform(l, u))

                print('Testing {}/{}, zoomlvl {},'.format(i+1, nsamples, z+1), 'params =', params)

                # Yield
                yield params

                # Track best ones
                if(self._curr_cost < min_cost):
                    min_cost = self._curr_cost
                    best_params = params
                    self._best = True
                else:
                    self._best = False

            # Zoom in around the best set of hyperparams
            new_range_types = []
            for best_param, rtype in zip(best_params, range_types):
                l = rtype[0]
                u = rtype[1]
                diff = u-l
                if(rtype[2] == 'log'): best_param = math.log10(best_param)
                new_l = best_param-(diff/zoomratio)
                new_u = best_param+(diff/zoomratio)
                new_range_types.append([new_l, new_u, rtype[2]])
            range_types = new_range_types


        # Print the best one
        print('\nSearch Finished')
        print('===============')
        print('Best params:', best_params)
        print('Best cost:', min_cost)


import os
import random
import multiprocessing as mp
from math import ceil
from contextlib import redirect_stdout

import numpy as np
import tqdm

from . import _shared_array
from ._regressor import Regressor
from ._classifier import Classifier
from . import decision_tree

class _RandomTree(decision_tree.DecisionTree):
    def __init__(self, input_size, output_size, num_features, feature_type=[], 
        max_depth=6, min_split=2, max_splits_eval=100, regression=False):
        # Initialize parent class
        super(_RandomTree, self).__init__(input_size, output_size, feature_type, max_depth,
            min_split, max_splits_eval, regression)

        # Select only a few random columns of the dataset for training
        self._cols_train = np.random.choice(input_size, num_features, replace=False)

        # Disable progress bar
        self._pbardis = True
        

class RandomForest(Classifier, Regressor):
    def __init__(self, input_size, output_size, feature_type=[], max_depth=6, min_split=2,
        max_splits_eval=100, regression=False):
        '''
        Parameters
        ----------
        input_size : int
            Size of input data or number of input features.
        output_size : int
            Number of categories or groups.
        feature_type : list
            List of string describing the type of feature for
            each column. Can be :code:`'continues'`, 
            :code:`'ranked'`, or :code:`'categorical'`.
        max_depth : int
            The maximum depth the trees can grow to.
        min_split : int
            The minimum number of data points a node should have to get 
            split.
        max_splits_eval : int
            The maximum number of split points to evaluate for an 
            attribute. If the number of candidate split points exceed
            this, :code:`max_splits_eval` split candidates will be
            randomly sampled from the candidates and only the sampled
            ones will be evaluated from finding the best split point.
        regression : bool
            If the tree is being trained on a regression problem.

        Raises
        ------
        InvalidFeatureType
            Invalid/Unknown feature type. Can only be :code:`'continues'`, 
            :code:`'ranked'`, or :code:`'categorical'`.
        '''
        # Save values
        self._input_size = input_size
        self._output_size = output_size
        self._ftype = feature_type
        self._max_depth = max_depth
        self._min_split = min_split
        self._regression = regression
        self._max_splits_eval = max_splits_eval

        # List to store trees in
        self._trees = []

        # Outputs
        self._output = None

    @property
    def _out_size(self):
        return self._output_size
    
    @property
    def trees(self):
        '''
        A list of decision trees used in the forest.
        '''
        return self._trees

    def train(self, inputs, outputs, num_trees=100, num_feature_bag=None):
        '''
        Trains the model on the training data.

        Parameters
        ----------
        training_data : numpy.array
            numpy array containing training data.
        targets : numpy.array
            numpy array containing training targets, corresponding to the training data.
        num_trees : int
            Number of trees to grow.
        num_feature_bag : int or None
            Number of random features to select when growing
            a tree. If :code:`None` (default), :code:`ceil(sqrt(input_size))`
            is chosen for classification and :code:`int(input_size/3)` for regression.

        Raises
        ------
        numpy.AxisError
            If output_size is less than two. Use :py:func:`pykitml.onehot` to change
            0/False to [1, 0] and 1/True to [0, 1] for binary classification.
        '''
        print('Training Model...')

        # Number of features to bag/choose for each tree
        if(num_feature_bag is None): 
            if(not self._regression):
                num_feature_bag = ceil(np.sqrt(self._input_size))
            else:
                num_feature_bag = int(self._input_size/3)
        
        def train_trees(input_q, ret_q, inputs_sh, inputs_shape, outputs_sh, outputs_shape):
            # Retrive numpy arrays from multiprocessing arrays
            inputs = _shared_array.shm_as_ndarray(inputs_sh, inputs_shape)
            output = _shared_array.shm_as_ndarray(outputs_sh, outputs_shape)
            
            # Supress print statements
            with redirect_stdout(open(os.devnull, 'w')):
                while(True):
                    # Get tree from input queue
                    try:
                        tree = input_q.get(block=False)
                    except mp.queues.Empty:
                        break

                    # Create bootstraped datset
                    indices = np.random.choice(inputs.shape[0], inputs.shape[0])
                    bootstrapped_inputs = inputs[indices]
                    bootstrapped_outputs = outputs[indices]

                    # Grow the tree
                    tree.train(bootstrapped_inputs, bootstrapped_outputs)  

                    # Put the trained tree in output queue
                    ret_q.put(tree)  

        # Create queues
        input_q = mp.Queue()
        ret_q = mp.Queue()

        # Initialize input queue
        for i in range(num_trees):
            # Create tree
            tree = _RandomTree(self._input_size, self._output_size, num_feature_bag, 
                self._ftype, self._max_depth, self._min_split, self._max_splits_eval,
                self._regression)
            # Put it in queue
            input_q.put(tree)

        # Create shared multiprocess array for inputs and outputs
        inputs_sh = _shared_array.ndarray_to_shm(inputs)
        outputs_sh = _shared_array.ndarray_to_shm(outputs)

        # Start multiprocess
        for i in range(os.cpu_count()):
            p = mp.Process(
                target=train_trees, args=(input_q, ret_q, 
                    inputs_sh, inputs.shape, outputs_sh, outputs.shape)
            )
            p.start()

        # Progress bar and append trained trees to list
        pbar = tqdm.tqdm(total=num_trees, ncols=80, unit='trees')
        
        while(len(self._trees) != num_trees):
            tree = ret_q.get()
            self._trees.append(tree)
            pbar.update()
        
        # Return if done
        pbar.close()

    def feed(self, input_data):
        # Loop through all the trees and total their outputs
        total = 0
        for tree in self._trees:
            tree.feed(input_data)
            total += tree.get_output()
        
        # Average
        self._output = total/len(self._trees)
        
    
    def get_output(self):
        return self._output.squeeze()

        




import pickle

'''
This module contains functions for saving and 
loading .pkl files
'''

def save(object_, file_name):
    '''
    Saves an object into a file.

    Parameters
    ----------
    object_ : object
        The object to save
    file_name : str
        The name of the file to save the object in.

    Raises
    ------
        OSError
            If the file cannot be created due to a system-related error.
    '''
    file = open(file_name, 'wb')
    pickle.dump(object_, file)
    file.close()


def load(file_name):    
    '''
    Loads an object from file.

    Parameters
    ----------
    file_name : str
        The name of the file to load the object from.

    Returns
    -------
    object
        The python object stored in the file.

    Raises
    ------
    FileNotFoundError
        If the file does not exist.
    '''
    file = open(file_name, 'rb')
    object_ = pickle.load(file)
    file.close()
    return object_



from copy import deepcopy

import numpy as np
import tqdm

from ._classifier import Classifier
from ._exceptions import _valid_list, InvalidDistributionType

def _gaussian(x, mean, std_dev):
    sqrt_2pi = np.sqrt(2*np.pi)
    return (1/(std_dev*sqrt_2pi))*np.exp(-0.5*(((x-mean)/std_dev)**2))

class NaiveBayes(Classifier):
    '''
    Implements Naive Bayes classifier.

    Note
    ----
    Consider using :class:`.GaussianNaiveBayes` if all of your
    features are continuous.
    '''

    def __init__(self, input_size, output_size, distributions, reg_param=1):
        '''
        Parameters
        ----------
        input_size : int
            Size of input data or number of input features.
        output_size: int
            Number of categories or groups.
        distribution : list
            List of strings describing the distribution to use 
            for each feature. Option are :code:`'gaussian'`, 
            :code:`'binomial'`, :code:`'multinomial'`.
        reg_param : int
            If a given class and feature value never occur together in the training data,
            then the frequency-based probability estimate will be zero.
            This is problematic because it will wipe out all information in the other 
            probabilities when they are multiplied.
            So, the probability will become :code:`log(reg_param)`.
            This is a way to regularize Naive Bayes classifier.
            See https://en.wikipedia.org/wiki/Naive_Bayes_classifier#Multinomial_naive_Bayes

        Raises
        ------
        InvalidDistributionType
            If invalid distribution. Can only be :code:`'gaussian'`, 
            :code:`'binomial'`, :code:`'multinomial'`.
        IndexError
            If the input_size does not match the length of distribution length.
        '''
        # Save info
        self._input_size = input_size
        self._output_size = output_size
        self._reg_param = reg_param

        self._dists = distributions

        # Check if given distributions are valid
        valid_dists = ['gaussian', 'binomial', 'multinomial']
        if not _valid_list(distributions, valid_dists):
            raise InvalidDistributionType

        # indices of categorical features
        self._categorical = [i for i, x in enumerate(distributions) if x!='gaussian']

        # for each class/group, p(class), std_dev, mean, freq and range
        self._pclass = np.zeros((output_size))
        self._mean = np.zeros((output_size, input_size))
        self._std_dev = np.zeros((output_size, input_size))
        self._freqp = [deepcopy([None]*input_size) for x in range(output_size)]
        self._max = None

        # Output
        self._output = None

    @property
    def _out_size(self):
        return self._output_size

    def feed(self, input_data):
        # Make sure array is 2D
        if(input_data.ndim == 1): input_data = np.array([input_data])

        # Set output to correct size
        self._output = np.zeros((input_data.shape[0], self._output_size))
        
        # Loop through each group and find probability
        for C in range(0, self._output_size):
            
            # Loop through each feature and multiply prod(p(xi|Ci))=p(x|Ci)
            # We are doing sum(log(p(xi|Ci))) instead of prod(p(xi|Ci)) to prevent
            # the product from going to zero due to very small numbers.
            # So the result will be log(p(x|Ci)) instead of p(x|Ci)
            p_xci = np.zeros((input_data.shape[0]))
            for x in range(0, self._input_size):
                # p(xi|Ci) if catgorical feature
                if(self._dists[x] != 'gaussian'):
                    p_xici = self._freqp[C][x][input_data[:, x].astype(int)]
                # p(xi|Ci) if continues feature
                else:    
                    p_xici = _gaussian(input_data[:, x], self._mean[C][x], self._std_dev[C][x])
                
                # log(p(x|Ci)) = sum(log(p(xi|Ci)))
                p_xci += np.log(p_xici)
            
            # log(p(Ci|x)) = log(p(Ci))+log(p(x|Ci))
            self._output[:, C] = np.log(self._pclass[C])+p_xci

    def get_output(self):
        return self._output.squeeze()

    def train(self, training_data, targets):
        '''
        Trains the model on the training data.

        Parameters
        ----------
        training_data : numpy.array
            numpy array containing training data.
        targets : numpy.array
            numpy array containing training targets, corresponding to the training data.

        Raises
        ------
        numpy.AxisError
            If output_size is less than two. Use :py:func:`pykitml.onehot` to change
            0/False to [1, 0] and 1/True to [0, 1] for binary classification.

        '''
        print('Training Model...')

        # Get max values for each feature
        self._max = np.amax(training_data, axis=0)

        # Loop through each group
        for group in tqdm.trange(0, self._output_size, ncols=80, unit='groups'):
            # Get examples for the group
            group_examples = self._get_group_examples(training_data, targets, group)
            # Get mean and standard deviation for the group examples
            self._get_mean_std(group, group_examples, training_data.shape[0])
            # Get frequency/probability for categorical 
            self._get_freq(group, group_examples)

    def _get_group_examples(self, training_data, targets, group):
        # Create one hot vector
        gvec = np.zeros((self._output_size))
        gvec[group] = 1

        # Get all example data for this group from training data
        indices = np.argwhere(np.all(targets == gvec, axis=1))
        group_examples = training_data[indices].squeeze()

        # return
        return group_examples

    def _get_mean_std(self, group, group_examples, tot_examples):
        # Get mean, std_dev and p(class)
        self._pclass[group] = group_examples.shape[0]/tot_examples
        self._mean[group] = np.mean(group_examples, axis=0)
        self._std_dev[group] = np.std(group_examples, axis=0)

    def _get_freq(self, group, group_examples):
        # Get frequencies
        for feature in self._categorical:
            # Get frequencies of each category in the feature for this
            # group example
            freqp = np.bincount(group_examples[:, feature].astype(int))
            freqp = freqp/group_examples.shape[0]
            # Replace with reg_param where p(xi|Ci) is zero
            # AKA regularization            
            freqp_reg = np.ones((int(self._max[feature])+1))*self._reg_param
            freqp_reg[0:freqp.shape[0]] = freqp
            self._freqp[group][feature] = freqp_reg


class GaussianNaiveBayes(NaiveBayes):
    def __init__(self, input_size, output_size):
        '''
        Parameters
        ----------
        input_size : int
            Size of input data or number of input features.
        output_size: int
            Number of categories or groups.
        '''
        # Save info
        self._input_size = input_size
        self._output_size = output_size

        # for each class/group, p(class), std_dev, mean, freq and range
        self._pclass = np.zeros((output_size))
        self._mean = np.zeros((output_size, input_size))
        self._std_dev = np.zeros((output_size, input_size))

        # Output
        self._output = None

    def feed(self, input_data):
        # Make sure array in 2D
        if(input_data.ndim == 1): input_data = np.array([input_data])

        # Set output to correct size
        self._output = np.zeros((input_data.shape[0], self._output_size))
        
        # Loop through each group and find probability
        for C in range(0, self._output_size):
            # Calculate p(xi|Ci)
            p_xici = _gaussian(input_data, self._mean[C], self._std_dev[C])
            
            # log(p(x|Ci)) = sum(log(p(xi|Ci)))
            p_xci = np.log(p_xici).sum(axis=1)
            
            # log(p(Ci|x)) = log(p(Ci))+log(p(x|Ci))
            self._output[:, C] = np.log(self._pclass[C])+p_xci

    def train(self, training_data, targets):
        print('Training Model...')

        # Loop through each group
        for group in tqdm.trange(0, self._output_size, ncols=80, unit='groups'):
            # Get examples for the group
            group_examples = self._get_group_examples(training_data, targets, group)
            # Get mean and standard deviation for the group examples
            self._get_mean_std(group, group_examples, training_data.shape[0])



from abc import ABC, abstractmethod

import numpy as np
import matplotlib.pyplot as plt

import tqdm

class MinimizeModel(ABC):
    '''
    Abstract base class for all models that use gradients and minimize a cost function.
    '''

    def train(self, training_data, targets, batch_size, epochs, optimizer,
            testing_data=None, testing_targets=None, testing_freq=1, decay_freq=1):
        '''
        Trains the model on the training data, after training is complete, you can call
        :py:func:`plot_performance` to plot performance graphs.

        Parameters
        ----------
        training_data : numpy.array
            numpy array containing training data.
        targets : numpy.array
            numpy array containing training targets, corresponding to the training data.
        batch_size : int
            Number of training examples to use in one epoch, or
            number of training examples to use to estimate the gradient.
        epochs : int
            Number of epochs the model should be trained for.
        optimizer : any Optimizer object
            See :ref:`optimizers`
        testing_data : numpy.array
            numpy array containing testing data.
        testing_targets : numpy.array
            numpy array containing testing targets, corresponding to the testing data.
        testing_freq : int
            How frequently the model should be tested, i.e the model will be tested
            after every :code:`testing_freq` epochs. You may want to increase this to reduce 
            training time.
        decay_freq : int
            How frequently the model should decay the learning rate. The learning rate
            will decay after every :code:`decay_freq` epochs.

        Raises
        ------
        ValueError
            If :code:`training_data`, :code:`targets`, :code:`testing_data` or 
            :code:`testing_targets` has invalid dimensions/shape. 
        '''
        print('Training Model...')

        # Dictionary for holding performance log
        self._performance_log = {}
        self._performance_log['epoch'] = []
        self._performance_log['cost_train'] = []
        self._performance_log['learning_rate'] = []
        if(testing_data is not None):
            self._performance_log['cost_test'] = []   

        self._init_train(batch_size)

        # Loop through each epoch
        pbar = tqdm.trange(0, epochs, ncols=80, unit='epochs')
        for epoch in pbar:
            total_gradient = self._get_batch_grad(epoch, batch_size, training_data, targets)

            # After completing a batch, average the total sum of gradients and tweak the parameters
            self._mparams = optimizer._optimize(self._mparams, total_gradient)

            # Decay the learning rate
            if((epoch+1) % decay_freq == 0): optimizer._decay()

            # Log and print performance
            if((epoch+1)%testing_freq == 0):
                # log epoch
                self._performance_log['epoch'].append(epoch+1)
                
                # log learning rate
                learning_rate = optimizer._learning_rate
                self._performance_log['learning_rate'].append(learning_rate)
                
                # get cost of the model on training data
                cost_train = self.cost(training_data, targets)
                pbar.set_postfix(cost=cost_train)
                self._performance_log['cost_train'].append(cost_train)
                
                # get cost of the model on testing data if it is provided
                if(testing_data is None): continue
                cost_test = self.cost(testing_data, testing_targets)
                self._performance_log['cost_test'].append(cost_test)

        # Close progress bar
        pbar.close()

    def _get_batch_grad(self, epoch, batch_size, training_data, targets):
        '''
        Calculates the sum of all the gradients fot the given batch

        Parameters
        ----------
        epoch : int
            Current epoch in the training process.
        batch_size : int
            Size of the batch.
        training_data : numpy.array
            numpy array containing training data.
        targets : numpy.array
            numpy array containing training targets, corresponding to the training data.
        '''
        # Total gradient for the batch 
        total_gradient = 0
        
        # feed the batch
        start_index = (epoch*batch_size)%training_data.shape[0]
        end_index = start_index+batch_size
        indices = np.arange(start_index, end_index) % training_data.shape[0]
        self.feed(training_data[indices])

        if(self.bptt):
            return self._backpropagate(None, targets[indices])
        else:
            # Loop through the batch
            for example in range(0, batch_size):
                # Add the calculated gradients to the total
                index = ((epoch*batch_size) + example)%training_data.shape[0]
                # Get gradient
                total_gradient += self._backpropagate(example, targets[index])

            # return the total
            return total_gradient/batch_size

    def plot_performance(self):
        '''
        Plots logged performance data after training. 
        Should be called after :py:func:`train`.

        Raises
        ------
        AttributeError
            If the model has not been trained, i.e :py:func:`train` has
            not been called before.
        IndexError
            If :py:func:`train` failed.
        '''
        graph = self._performance_log

        # Window title
        plt.figure('Performance graph', figsize=(10, 7))

        # First subplot
        plt.subplot(2, 1, 1)

        # Plot average cost vs epochs on training data
        plt.plot(graph['epoch'], graph['cost_train'], label='Training data')
        
        # Plot average cost on testing data
        if('cost_test' in graph.keys()):
            plt.plot(graph['epoch'], graph['cost_test'], label='Test data')

        # Axis labels
        plt.ylabel('Average cost')
        plt.xlabel('No. of epochs')
        plt.legend()

        # Second subplot
        plt.subplot(2, 1, 2)

        # Plot learning rate vs epochs
        plt.plot(graph['epoch'], graph['learning_rate'], label='Learning Rate')

        # Axis labels
        plt.ylabel('Learning Rate')
        plt.xlabel('No. of epochs')
        plt.legend()        

        # Show the plot
        plt.show()

    def _init_train(self, batch_size):
        '''
        This function will be called before training starts.
        Override this function to do initialization.
        
        Parameters
        ----------
        batch_size : int
            Size of the batch.       
        '''
        pass

    def cost(self, testing_data, testing_targets):
        '''
        Tests the average cost of the model on the testing data passed to the
        function.

        Parameters
        ----------
        testing_data : numpy.array
            numpy array containing testing data.
        testing_targets : numpy.array
            numpy array containing testing targets, corresponding to the testing data.
        
        Returns
        -------
        cost : float
            The average cost of the model over the testing data.

        Raises
        ------
        ValueError
            If :code:`testing_data` or :code:`testing_tagets` has invalid dimensions/shape.      
        '''
        self._on_test_start()

        # Evalute over all the testing data and get outputs
        self.feed(testing_data)
        output_targets = self.get_output()

        # Calculate cost
        cost = np.sum(self._cost_function(output_targets, testing_targets))
        # Add regularization
        cost += self._get_norm_weights()
        
        # Average the cost
        cost = cost/testing_data.shape[0]

        self._on_test_end()

        # return cost
        return round(cost, 2)

    def _on_test_start(self):
        '''
        This method will be called before testing, 
        override this method needed.
        '''
        pass

    def _on_test_end(self):
        '''
        This method will be called after testing, 
        override this method needed.
        '''
        pass

    @property
    @abstractmethod
    def _mparams(self):
        '''
        @property method to get model parameters
        '''
        pass
    
    @_mparams.setter
    @abstractmethod
    def _mparams(self, mparams):
        '''
        @params.setter method to set model parameters
        '''
        pass

    @property
    @abstractmethod
    def bptt(self):
        '''
        Return True if the model requires BPTT
        (Backpropagation through time), otherwise return false
        '''
        pass

    @abstractmethod
    def feed(self, input_data):
        '''
        Accepts input array and feeds it to the model.

        Parameters
        ----------
        input_data : numpy.array
            The input to feed the model.

        Raises
        ------
        ValueError
            If the input data has invalid dimensions/shape.      

        Note
        ----
        This function only feeds the input data, to get the output after calling this
        function use :py:func:`get_output` or :py:func:`get_output_onehot`
        '''
        pass

    @abstractmethod
    def get_output(self):
        '''
        Returns the output activations of the model.

        Returns
        -------
        numpy.array
            The output activations.
        '''
        pass

    @abstractmethod
    def _backpropagate(self, index, targets):
        '''
        This function calculates gradient of the cost function w.r.t all weights and 
        biases of the model by backpropagating the error through the model.

        Parameters
        ----------
        index : int
            Index of the example in the batch that was fed using feed.
        target : numpy.array
            The correct activations that the output layer should have.

        Returns
        -------
        gradient: numpy.array
            Numpy object array containing gradients, has same shape as self._mparams

        Raises
        ------
        ValueError
            If the input data has invalid dimensions/shape.

        Note
        ----
        You have to call :py:func:`~feed` before you call this function.
        ''' 
        pass

    @property
    @abstractmethod
    def _cost_function(self):
        '''
        Return the cost function used in the model.
        '''
        pass

    @abstractmethod
    def _get_norm_weights(self):
        '''
        Return the norm of all the regularized parameters of the models
        multiplied by the regularization parameter.
        '''
        pass

    


'''
This module contains helper functions to download and load MNIST and MNIST like datasets.   
'''

# ============================================================  
# = Forked from: https://github.com/hsjeong5/MNIST-for-Numpy =
# = Modified with minor changes                              =
# ============================================================

import gzip
import pickle
import os
from urllib import request

import numpy as np

from .. import pklhandler

def get(type = 'classic'):
    '''
    Downloads the MNIST dataset and saves it as a pickle file, `mnist.pkl`.

    Parameters
    ----------
    type : str
        The type of MNIST dataset to download.

        - 'classic' : Downloads the classic hanwritten digits dataset from http://yann.lecun.com/exdb/mnist/
        - 'fashion' : Downloads fashion MNIST from https://github.com/zalandoresearch/fashion-mnist


    Raises
    ------
        urllib.error.URLError
            If internet connection is not available or the URL is not accessible.
        OSError
            If the file cannot be created due to a system-related error.
        KeyError
            If invalid/unknown type.

    Note
    ----
    You only need to call this method once, i.e, after the dataset has been 
    downloaded and you have the `mnist.pkl` file, you don't need to call this method again.
    '''
    # dict of URLs containing MNIST like datasets
    type_URLs = {'classic':'http://yann.lecun.com/exdb/mnist/',
            'fashion':'http://fashion-mnist.s3-website.eu-central-1.amazonaws.com/'
    }

    # MNIST files to download
    filename = [
        ['training_images','train-images-idx3-ubyte.gz'],
        ['test_images','t10k-images-idx3-ubyte.gz'],
        ['training_labels','train-labels-idx1-ubyte.gz'],
        ['test_labels','t10k-labels-idx1-ubyte.gz']
    ]

    def download_mnist():
        # Download .gz files
        base_url = type_URLs[type]
        for name in filename:
            print('Downloading '+name[1]+'...')
            request.urlretrieve(base_url+name[1], name[1])
        print('Download complete.')

    def save_mnist():
        # Read .gz files and put them in a numpy array and save it as a pkl file
        mnist = {}
        for name in filename[:2]:
            with gzip.open(name[1], 'rb') as f:
                mnist[name[0]] = np.frombuffer(f.read(), np.uint8, offset=16).reshape(-1,28*28)
        for name in filename[-2:]:
            with gzip.open(name[1], 'rb') as f:
                mnist[name[0]] = np.frombuffer(f.read(), np.uint8, offset=8)
        pklhandler.save(mnist, 'mnist.pkl')
        print('Save complete.')

    def clean():
        # Remove unnecessary files
        os.remove('train-images-idx3-ubyte.gz')
        os.remove('t10k-images-idx3-ubyte.gz')
        os.remove('train-labels-idx1-ubyte.gz')
        os.remove('t10k-labels-idx1-ubyte.gz')
        print('Deleted unnecessary files.')

    download_mnist()
    save_mnist()
    clean()


def load():
    '''
    Loads MNIST dataset from saved pickle file `mnist.pkl` to numpy arrays.

    Returns
    -------
        training_data : numpy.array
            60,000x784 numpy array, each row contains flattened version of training images.
        training_targets : numpy.array
            60,000x10 numpy array that contains one hot target array of the corresponding 
            training images.
        testing_data : numpy.array
            10,000x784 numpy array, each row contains flattened version of test images.
        testing_targets : numpy.array
            10,000x10 numpy array that contains one hot target array of the corresponding
            test images.

    Raises
    ------
        FileNotFoundError
            If `mnist.pkl` file does not exist, i.e, if the dataset was not downloaded and
            saved using the :py:func:`~get` method. 
    '''
    mnist = pklhandler.load('mnist.pkl')
    # Normalize data
    training_data = mnist['training_images']/255
    testing_data = mnist['test_images']/255
    # Create one-hot target array for training labels
    training_targets = np.zeros((60000, 10))
    training_targets[np.arange(60000), mnist['training_labels']] = 1
    # Create one-hot target array for testing labels
    testing_targets = np.zeros((10000, 10))
    testing_targets[np.arange(10000), mnist['test_labels']] = 1
    # return the data
    return training_data, training_targets, testing_data, testing_targets


if __name__ == '__main__':
    get()
    


from abc import ABC, abstractmethod
import itertools

import numpy as np
import matplotlib.pyplot as plt

import tqdm

from . import _heatmap
from . import preprocessing

class Regressor(ABC):
    '''
    Mix-in class for Regression models.
    '''

    @abstractmethod
    def get_output(self):
        '''
        Returns the output activations of the model.

        Returns
        -------
        numpy.array
            The output activations.
        '''
        pass

    @abstractmethod
    def feed(self, input_data):
        '''
        Accepts input array and feeds it to the model.

        Parameters
        ----------
        input_data : numpy.array
            The input to feed the model.

        Raises
        ------
        ValueError
            If the input data has invalid dimensions/shape.      

        Note
        ----
        This function only feeds the input data, to get the output after calling this
        function use :py:func:`get_output` or :py:func:`get_output_onehot`
        '''
        pass

    @property
    @abstractmethod
    def _out_size(self):
        '''
        Returns number of nodes/neurons in the output layer.
        '''
        pass

    def r2score(self, testing_data, testing_targets):
        '''
        Return R-squared or coefficient of determination value.
        
        Parameters
        ----------
        testing_data : numpy.array
            numpy array containing testing data.
        testing_targets : numpy.array
            numpy array containing testing targets, corresponding to the testing data.
        
        Returns
        -------
        r2score : float
            The average cost of the model over the testing data.

        Raises
        ------
        ValueError
            If :code:`testing_data` or :code:`testing_tagets` has invalid dimensions/shape.      
        '''
        self.feed(testing_data)
        output = self.get_output()

        error = ((output-testing_targets)**2).sum()
        var = ((testing_targets-testing_targets.mean(axis=0)) ** 2).sum()

        return 1-error/var




'''
pykitml (Python Kit for Machine Learning),
    pykitml Machine Learning library.
Copyrights(c) Vishnu Shankar
MIT License (See LICENSE file)
https://github.com/RainingComputers
'''

from .network import *
from .linear_regression import *
from .logistic_regression import *
from .svm import *
from .naive_bayes import *
from .decision_tree import *
from .random_forest import *
from .nearest_neighbor import *
from .pca import *
from .kmeans_clustering import *
from .pklhandler import *
from .normalize import *
from .optimizers import *
from .preprocessing import *
from .cross_val import *
from .lstm import *
from .smote import *
from .random_search import *
from .fceux import *
from .dqn import *

from . import testing



import numpy as np
import tqdm
import matplotlib.pyplot as plt

from . import _functions
from . import normalize

def kmeans(training_data, nclusters, max_iter=1000, trials=50):
    '''
    Identifies cluster centres on training data using k-means.

    Parameters
    ----------
    training_data : numpy.array
        Numpy array containing training data.
    nclusters : int
        Number of cluster to find.
    max_iter : int
        Maximum number of iterations to run per trial.
    trials : int
        Number of times k-means should run, each with different
        random initialization.

    Returns
    -------
    clusters : numpy.array
        Numpy array containing cluster centres.
    cost : numpy.array
        The cost of converged cluster centres.

    '''
    
    # Keep track of trial with the least cost
    min_cost = np.float('infinity')
    distances = None
    clusters_min_cost = None
    clusters = None

    # Keep log of maximum number of iterations for convergence
    max_iter_log = 0
    
    pbar = tqdm.trange(0, trials, ncols=80, unit='trials')
    for trial in pbar:
        # Use kmeans++ to initialize cluster centres
        clusters = np.zeros((nclusters, training_data.shape[1]))
        
        # First cluster centre is random
        index = np.random.randint(training_data.shape[0], size=1)
        clusters[0] = training_data[index]
        
        # Loop for rest of cluster centres
        for i in range(1, nclusters):
            # Calculate distance between every data point and previous cluster centre
            prev_cluster_dists = _functions.pdist(clusters[i-1], training_data).squeeze()
            # Normalize distances
            prev_cluster_dists = prev_cluster_dists/prev_cluster_dists.sum()
            
            # Sample index with probability distribution proportional to distances
            index = np.random.choice(training_data.shape[0], 1, p=prev_cluster_dists)
            
            # Assign next cluster centre
            clusters[i] = training_data[index]
        
        # Start kmeans, Keep looping and moving the cluster points to mean
        for iteration in range(max_iter):
            new_clusters = np.zeros((nclusters, training_data.shape[1]))

            # Calculate distances between clusters and every point in training data 
            distances = _functions.pdist(training_data, clusters)

            # Assign clusters index to each data point
            cluster_assignments = np.argmin(distances, axis=1)

            # Move cluster by taking mean of all the points assigned to that cluster
            for i in range(nclusters):
                cluster_points = training_data[cluster_assignments==i]
                if(cluster_points.shape[0] == 0): continue
                new_clusters[i] = np.mean(cluster_points, axis=0)
            
            # Check for convergence
            if(np.abs(new_clusters-clusters)==0).all(): break

            # Assign new clusters
            clusters = new_clusters

        # Select cluster centres with least cost
        cost = np.mean(np.min(distances, axis=1))
        if(cost < min_cost):
            clusters_min_cost = clusters
            min_cost = cost

        # Update maximum iterations for convergence
        if(iteration > max_iter_log):
            max_iter_log = iteration

        # Update progress bar
        pbar.set_postfix(cost=min_cost, max_it=max_iter_log)

    return clusters_min_cost, min_cost



from copy import deepcopy

import numpy as np

from ._minimize_model import MinimizeModel
from ._classifier import Classifier
from ._regressor import Regressor

from . import _functions
from ._functions import tanh, sigmoid, tanh_prime, sigmoid_prime     
from ._functions import mse, mse_prime

__all__ = ['LSTM']

# Enums
Wz, Wi, Wf, Wo, Rz, Ri, Rf, Ro, Bz, Bi, Bf, Bo, OUT_W, OUT_B = list(range(14)) 

class LSTM(MinimizeModel, Classifier, Regressor):
    '''
    This class implements LSTM Network.
    '''
    def __init__(self, layer_sizes, output_activ_func='softmax', cost_func='cross_entropy'):
        '''
        Parameters
        ----------
        layer_sizes : list
            A list of integers describing the number of layers and the number of neurons in each
            layer. For e.g. :code:`[784, 100, 100, 10]` describes a network with one input
            layer having 784 neurons, two hidden LSTM layers having 100 neurons each and a 
            dense output layer with 10 neurons.
        output_activ_function : str
            Activation function to use for dense output layer.
            List of available activation functions:
            :code:`leakyrelu`, :code:`relu`, :code:`softmax`, :code:`tanh`, :code:`sigmoid`, :code:`identity`.
        cost_function : str
            List of available cost functions:
            :code:`mse` (Mean Squared Error), :code:`cross_entropy` (Cross Entropy), :code:`huber` (Huber loss).
        '''
        self.t = 1
        self._lsizes = layer_sizes
        self._nlayers = len(layer_sizes)
        self._bsize = None
        self._outputs = None
        self._batch_count = 0

        # Initialize output activation function and cost function
        self._output_activ_func = getattr(_functions, output_activ_func)
        self._output_activ_func_prime = getattr(_functions, output_activ_func+'_prime')
        self._cost_func = getattr(_functions, cost_func)
        self._cost_func_prime = getattr(_functions, cost_func+'_prime')

        # Initialize LSTM gates with list for each layer
        [self.z_ws, self.z, self.i_ws, self.i, self.f_ws, self.f, self.c, 
            self.o_ws, self.o, self.y, self.last_y, self.last_c] = \
                [[np.array([])]*(self._nlayers-1) for i in range(12)]
        
        # Initialize dense output layer activations
        self.out_ws = None
        self.out_a = None 

        # Initialize LSTM weights
        [w_z, w_i, w_f, w_o, r_z, r_i, r_f, r_o, b_z, 
            b_i, b_f, b_o] = [[np.array([])]*(self._nlayers-1) for i in range(12)]
        # Loop through each layer and initialize random weights and b_z
        for l in range(1, self._nlayers-1):
            lsize = layer_sizes[l]
            input_lsize = layer_sizes[l-1]
            eps = np.sqrt(6 / (lsize + input_lsize))
            w_z[l], w_i[l], w_f[l], w_o[l] = [np.random.uniform(-eps, eps, (lsize, input_lsize)) for i in range(4)]
            r_z[l], r_i[l], r_f[l], r_o[l] = [np.random.uniform(-eps, eps, (lsize, lsize)) for i in range(4)]
            b_z[l], b_i[l], b_f[l], b_o[l] = [np.random.uniform(-eps, eps, (lsize)) for i in range(4)]
        
        # Initialize dense output layer weights
        lsize = layer_sizes[-1]
        input_lsize = layer_sizes[-2]
        eps = np.sqrt(6 / (lsize + input_lsize))
        out_w = np.random.uniform(-eps, eps, (lsize, input_lsize))
        out_b = np.random.uniform(-eps, eps, (lsize))

        # Put parameters in numpy dtype=object array
        self._params = np.array(
            [np.array(w_z, dtype=object), np.array(w_i, dtype=object),
                np.array(w_f, dtype=object), np.array(w_o, dtype=object),
                np.array(r_z, dtype=object), np.array(r_i, dtype=object),
                np.array(r_f, dtype=object), np.array(r_o, dtype=object),
                np.array(b_z, dtype=object), np.array(b_i, dtype=object),
                np.array(b_f, dtype=object), np.array(b_o, dtype=object),
                np.array([out_w], dtype=object),  np.array([out_b], dtype=object)],
            dtype=object
        )  

    @property
    def _mparams(self):
        return self._params
    
    @_mparams.setter
    def _mparams(self, mparams):
        self._params = mparams

    @property
    def _cost_function(self):
        return self._cost_func

    @property
    def _out_size(self):
        return self._lsizes[-1]

    @property
    def nlayers(self):
        '''
        The number of layers in the network.
        '''      
        return self._nlayers

    def _init_gates(self, batch_size):
        self.y[0] = np.zeros((batch_size+2, self._lsizes[0]))

        self.out_ws = np.zeros((batch_size+2, self._lsizes[-1]))
        self.out_a = np.zeros((batch_size+2, self._lsizes[-1]))

        self._bsize = batch_size
        for l in range(1, self._nlayers-1): 
            [self.z_ws[l], self.z[l], self.i_ws[l], self.i[l], self.f_ws[l], 
                self.f[l], self.c[l], self.o_ws[l], self.o[l], 
                self.y[l]] = [np.zeros((batch_size+2, self._lsizes[l])) for i in range(10)]
            
            self.last_y[l] = np.zeros((self._lsizes[l]))
            self.last_c[l] = np.zeros((self._lsizes[l]))

    def _reset_state(self):
        for l in range(1, self._nlayers-1): 
            self.last_y[l] = np.zeros((self._lsizes[l]))
            self.last_c[l] = np.zeros((self._lsizes[l]))

    def reset(self):
        '''
        Resets the hidden state.
        '''
        self.t = 1
        self._reset_state()

    def feed(self, input_data):
        self._outputs = np.zeros((input_data.shape[0], self._out_size))
        if(input_data.ndim == 1):
            # For one input example
            self._feed_ndim1(input_data)
            self._outputs = self.out_a[self.t-1]
        else:
            # For multiple input example
            for i in range(input_data.shape[0]):
                self._feed_ndim1(input_data[i])
                self._outputs[i] = self.out_a[self.t-1]

    def _feed_ndim1(self, input_data):
        if(self.t > self._bsize): self.t = 1
        t = self.t

        self.y[0][t] = input_data
        for l in range(1, self._nlayers-1):
            # Block input
            self.z_ws[l][t] = self.y[l-1][t]@self._params[Wz][l].T + self.last_y[l]@self._params[Rz][l].T + self._params[Bz][l]
            self.z[l][t] = tanh(self.z_ws[l][t])
            # Input gate
            self.i_ws[l][t] = self.y[l-1][t]@self._params[Wi][l].T + self.last_y[l]@self._params[Ri][l].T + self._params[Bi][l]
            self.i[l][t] = sigmoid(self.i_ws[l][t])
            # Forget gate
            self.f_ws[l][t] = self.y[l-1][t]@self._params[Wf][l].T + self.last_y[l]@self._params[Rf][l].T + self._params[Bf][l]
            self.f[l][t] = sigmoid(self.f_ws[l][t])
            # Output gate
            self.o_ws[l][t] = self.y[l-1][t]@self._params[Wo][l].T + self.last_y[l]@self._params[Ro][l].T + self._params[Bo][l]
            self.o[l][t] = sigmoid(self.o_ws[l][t])
            # Cell state
            self.c[l][t] = self.z[l][t]*self.i[l][t] + self.last_c[l]*self.f[l][t]
            # Block output
            self.y[l][t] = tanh(self.c[l][t])*self.o[l][t]

            # Update previous state
            self.last_y[l] = self.y[l][t]
            self.last_c[l] = self.c[l][t]

        # Feed through the last dense layer
        self.out_ws[t] = self.last_y[-1]@self._params[OUT_W][0].T + self._params[OUT_B][0]
        self.out_a[t] = self._output_activ_func(self.out_ws[t])

        self.t += 1

    def get_output(self):
        return self._outputs.squeeze()

    def _init_train(self, batch_size):
        # Initialize LSTM gates
        self._init_gates(batch_size)

    def _backpropagate(self, batch_size, targets):
        # Initialize LSTM gates' gradients
        [dz, di, df, dc, do, dy] = [[np.array([])]*(self._nlayers-1) for i in range(6)]
        for l in range(1, self._nlayers-1): 
            [dz[l], di[l], df[l], dc[l], do[l], dy[l]] = \
                [np.zeros((self._bsize+2, self._lsizes[l])) for i in range(6)]              

        # Initialize LSTM weights' gradients
        [dw_z, dw_i, dw_f, dw_o, dr_z, dr_i, dr_f, dr_o, db_z, 
            db_i, db_f, db_o] = [[np.array([])]*(self._nlayers-1) for i in range(12)]
        for l in range(1, self._nlayers-1):
            lsize = self._lsizes[l]
            input_lsize = self._lsizes[l-1]
            dw_z[l], dw_i[l], dw_f[l], dw_o[l] = [np.zeros((lsize, input_lsize)) for i in range(4)]
            dr_z[l], dr_i[l], dr_f[l], dr_o[l] = [np.zeros((lsize, lsize)) for i in range(4)]
            db_z[l], db_i[l], db_f[l], db_o[l] = [np.zeros((lsize)) for i in range(4)]

        # Initialize dense layer weights' gradients
        lsize = self._lsizes[-1]
        input_lsize = self._lsizes[-2]
        dout_w = np.zeros((lsize, input_lsize))
        dout_b = np.zeros((lsize))
        dcost = np.zeros((lsize))

        def calc_delta_t(l, t):
            if(l == self._nlayers-2):
                return (dcost*dout_a)@self._params[OUT_W][0]
            else:
                return dz[l+1][t]@self._params[Wz][l+1] + di[l+1][t]@self._params[Wi][l+1] \
                    + df[l+1][t]@self._params[Wf][l+1] + do[l+1][t]@self._params[Wo][l+1] 

        for t in range(self._bsize, 0, -1):
            dout_a = self._output_activ_func_prime(self.out_ws[t], self.out_a[t])
            dcost = self._cost_func_prime(self.out_a[t], targets[t-1])
            dout_b += dcost * dout_a
            dout_w += np.multiply.outer(dout_b, self.y[-1][t])

            for l in range(self._nlayers-2, 0, -1):
                dy[l][t] = calc_delta_t(l, t) + dz[l][t+1]@self._params[Rz][l] + di[l][t+1]@self._params[Ri][l]\
                    + df[l][t+1]@self._params[Rf][l] + do[l][t+1]@self._params[Ro][l]
                do[l][t] = dy[l][t] * tanh(self.c[l][t]) * sigmoid_prime(self.o_ws[l][t], self.o[l][t])
                dc[l][t] = dy[l][t] * self.o[l][t] * (1-(tanh(self.c[l][t])**2)) + dc[l][t+1]*self.f[l][t+1]
                df[l][t] = dc[l][t] * self.c[l][t-1] * sigmoid_prime(self.f_ws[l][t], self.f[l][t])
                di[l][t] = dc[l][t] * self.z[l][t] * sigmoid_prime(self.i_ws[l][t], self.i[l][t])
                dz[l][t] = dc[l][t] * self.i[l][t] * tanh_prime(self.z_ws[l][t], self.z[l][t])
                dw_z[l] += np.multiply.outer(dz[l][t], self.y[l-1][t])
                dw_i[l] += np.multiply.outer(di[l][t], self.y[l-1][t])
                dw_f[l] += np.multiply.outer(df[l][t], self.y[l-1][t])
                dw_o[l] += np.multiply.outer(do[l][t], self.y[l-1][t])
                dr_z[l] += np.multiply.outer(dz[l][t+1], self.y[l][t])
                dr_i[l] += np.multiply.outer(di[l][t+1], self.y[l][t])
                dr_f[l] += np.multiply.outer(df[l][t+1], self.y[l][t])
                dr_o[l] += np.multiply.outer(do[l][t+1], self.y[l][t])
                db_z[l] += dz[l][t]
                db_i[l] += di[l][t]
                db_f[l] += df[l][t]
                db_o[l] += do[l][t]

        return np.array(
            [np.array(dw_z, dtype=object), np.array(dw_i, dtype=object),
                np.array(dw_f, dtype=object), np.array(dw_o, dtype=object),
                np.array(dr_z, dtype=object), np.array(dr_i, dtype=object),
                np.array(dr_f, dtype=object), np.array(dr_o, dtype=object),
                np.array(db_z, dtype=object), np.array(db_i, dtype=object),
                np.array(db_f, dtype=object), np.array(db_o, dtype=object),
                np.array([dout_w], dtype=object), np.array([dout_b], dtype=object)
                ],
            dtype=object
        )  

    def _on_test_start(self):
        self._saved_state = (deepcopy(self.last_c), deepcopy(self.last_y))
        self._reset_state()

    def _on_test_end(self):
        self.last_c, self.last_y = self._saved_state

    @property
    def bptt(self):
        return True

    def _get_norm_weights(self):
        return 0




import random
from abc import ABC, abstractmethod
from collections import deque

import numpy as np
import matplotlib.pyplot as plt

import tqdm

from .network import NeuralNetwork
from .pklhandler import save, load
from ._functions import huber

'''
    Ref: https://github.com/keon/deep-q-learning/blob/master/ddqn.py
'''

__all__ = ['DQNAgent', 'Environment']

class ReplayBuffer:
    def __init__(self, state_size, buffer_size):
        self._size = buffer_size
        self._ptr = 0
        self._len = 0
        
        self._states = np.zeros((buffer_size, state_size))
        self._actions = np.zeros((buffer_size), dtype=int)
        self._rewards = np.zeros((buffer_size))
        self._next_states = np.zeros((buffer_size, state_size))
        self._done = np.zeros((buffer_size), dtype=bool)

    def __len__(self):
        return self._len

    def append(self, state, action, reward, next_state, done):
        if(self._len != self._size): self._len += 1
        if(self._ptr == self._size): self._ptr = 0

        ptr = self._ptr
        self._states[ptr] = state
        self._actions[ptr] = action
        self._rewards[ptr] = reward
        self._next_states[ptr] = next_state
        self._done[ptr] = done

        self._ptr += 1

    def sample(self, batch_size):
        indices = np.random.choice(self._len, batch_size)

        return (self._states[indices], self._actions[indices],
            self._rewards[indices], self._next_states[indices],
            self._done[indices])

class DQNAgent:
    '''
    This class implements Double Deep Q Learning for RL. Use a Neural Network
    with huber loss for predicting Q values.
    '''

    def __init__(self, layer_sizes, mem_size=10000):
        '''
        Parameters
        ----------
        layer_sizes : list
            A list of integers describing the number of layers and the number of neurons in each
            layer. For e.g. :code:`[784, 100, 100, 10]` describes a network with one input
            layer having 784 neurons, two hidden LSTM layers having 100 neurons each and a 
            dense output layer with 10 neurons.
        mem_size : int
            The size for storing experience replay.       
        '''
        self._model = NeuralNetwork(layer_sizes, config='leakyrelu-identity-huber')
        self._target_model = NeuralNetwork(layer_sizes, config='leakyrelu-identity-huber')
        self._target_model._mparams = np.array(self._model._mparams, copy=True)

        self._act_size = layer_sizes[-1]
        self._state_size = layer_sizes[0]

        self._memory = ReplayBuffer(layer_sizes[0], mem_size)

        self._save_freq = 0
        self._filename = ''
    
    def _train_minibatch(self, batch_size, optimizer, discount_factor):
        if(batch_size > len(self._memory)): return 0

        # Sample a mini batch
        states, actions, rewards, next_states, done = self._memory.sample(batch_size)
        
        # Calculate target qvals
        self._target_model.feed(next_states)
        target_qvals = np.where(done,
            rewards,
            rewards + discount_factor*np.amax(self._target_model.get_output(), axis=1)
        )

        # Train
        cost = 0
        for i in range(batch_size):
            self._model.feed(np.array([states[i]]))
            target = np.array(self._model.get_output(), copy=True)
            cost += huber(target[actions[i]], target_qvals[i])
            target[actions[i]] = target_qvals[i]
            grad = self._model._backpropagate(0, target)
            self._model._mparams = optimizer._optimize(self._model._mparams, grad)
                
        return cost/batch_size

    def train(self, env, nepisodes, optimizer, batch_size=64, render=False,
        update_freq = 1, explr_rate=1, explr_min=0.01, explr_decay=0.99, disc_factor=0.95):
        '''
        Trains the agent on the given environment using deep Q learning.

        Parameters
        ----------
        env : obj
            Object that represents the environment. See :ref:`environment`
        nepisodes : int
            Number of episodes to train for.
        optimizer : any Optimizer object
            See :ref:`optimizers`
        batch_size : int
            How many samples from replay experience to train on on each step.
        render : bool
            If set to true, will call the :code:`render()` method in :code:`env`
            object.
        update_freq : int
            How often to update the target model in episodes. 
        explr_rate : float
            Initial exploration rate. Higher the exploration rate,
            agent will take more random actions.
        explr_min : float
            Minimum exploration rate.
        explr_decay : float
            Multiplication factor for reducing exploration rate after
            each episode.
        disc_factor : float
            Discount factor, value of future rewards.
        '''

        self._reward_graph = []
        self._cost_graph = []

        pbar = tqdm.trange(0, nepisodes, ncols=80, unit='episodes')
        for e in pbar:
            # Play an episode in the environment
            done = False
            state = env.reset()
            total_rewards = 0
            total_steps = 0
            total_cost = 0
            while not done:
                if(random.random() <= explr_rate):
                    action = random.randint(0, self._act_size-1)
                else:
                    self._model.feed(state)
                    qval = self._model.get_output()
                    action = np.argmax(qval)

                next_state, reward, done = env.step(action)
                if(render): env.render()

                # Track metrics
                total_rewards += reward
                total_steps += 1
                
                # Store in replay memory
                self._memory.append(state, action, reward, next_state, done)
                state = next_state

                # Sample mini batch and train
                cost = self._train_minibatch(batch_size, optimizer, disc_factor)
                total_cost += cost

            # Decrase exploration rate
            if(explr_rate > explr_min): explr_rate*=explr_decay

            # Update target model
            if(e%update_freq == 0):
                self._target_model._mparams = np.array(self._model._mparams, copy=True)

            # Update progress bar
            pbar.set_postfix(steps=total_steps, rewards=total_rewards)
            self._reward_graph.append(total_rewards)
            self._cost_graph.append(total_cost/total_steps)

            # Save agent
            if(self._save_freq != 0):
                if((e+1)%self._save_freq == 0): 
                    self.save(self._filename+'_episode'+str(e+1)+'.pkl')

        env.close()

    def set_save_freq(self, freq, name):
        '''
        Saves the agent model every :code:`freq` episodes during training.

        Parameters
        ----------
        freq : int
            Saving frequency in episodes.
        name : str
            Name of the file.
        '''
        self._save_freq = freq
        self._filename = name

    def plot_performance(self, N=30):
        '''
        Plots logged performance data after training. 
        Should be called after :py:func:`train`.

        Parameters
        ----------
        N : int
            How many points to take for the running mean.

        Raises
        ------
        AttributeError
            If the model has not been trained, i.e :py:func:`train` has
            not been called before.        
        '''
        def running_mean(x, N):
            cumsum = np.cumsum(np.insert(x, 0, 0)) 
            return (cumsum[N:] - cumsum[:-N]) / float(N)

        plt.figure('Performance graph', figsize=(10, 7))

        plt.subplot(2, 1, 1)
        plt.ylabel('Reward')
        plt.plot(running_mean(np.array(self._reward_graph), N))
        
        plt.subplot(2, 1, 2)
        plt.ylabel('Cost')
        plt.xlabel('Episode')
        plt.plot(running_mean(np.array(self._cost_graph), N))

        plt.show()

    def exploit(self, env, render=False):
        '''
        Exploit the trained model to make decision. No training occurs.
        Use to demo the trained agent.
        
        Parameters
        ----------
        env : obj
            Object that represents the environment. See :ref:`environment`
        render : bool
            If set to true, will call the :code:`render()` method in :code:`env`
        '''
        while True:
            done = False
            state = env.reset()
            while not done:
                self._model.feed(state)
                qval = self._model.get_output()
                action = np.argmax(qval)

                next_state, reward, done = env.step(action)
                if(render): env.render()

                state = next_state

    def save(self, file_name):
        save((self._model, self._reward_graph, self._cost_graph), file_name)

    def load(self, file_name):
        self._model, self._reward_graph, self._cost_graph = load(file_name)

class Environment(ABC):
    @abstractmethod
    def reset(self):
        '''
        Resets the environment and returns initial state.

        Returns
        -------
        initial_state : np.array
            The initial state of the environment as a numpy array.
        '''
        pass

    @abstractmethod
    def step(self, action):
        '''
        Performas given action, modifies current state to next state, 
        returns next state, reward and a bool to tell weather the episode has
        terminated.

        Returns
        -------
        next_state : np.array
            The next state of the environment as a numpy array.
        reward : int
            An integer telling the agent how well it performed.
        done : bool
            Flag to tell weather the environment has reached a terminal
            state.
        '''
        pass

    @abstractmethod
    def close(self):
        '''
        Called after training is completely, properly closes/exists
        the environment.
        '''
        pass

    def render(self):
        '''
        Method to render the environment, show a visual representation
        of the environment.
        '''
        pass



import os
from urllib import request

import numpy as np

from .. import pklhandler

'''
This module contains helper functions to download and load
the heart disease dataset. 
'''

def get():
    '''
    Downloads heartdisease dataset from
    https://archive.ics.uci.edu/ml/datasets/Heart+Disease
    and saves it as a pkl file `heartdisease.pkl`.

    Raises
    ------
        urllib.error.URLError
            If internet connection is not available or the URL is not accessible.
        OSError
            If the file cannot be created due to a system-related error.
        KeyError
            If invalid/unknown type.

    Note
    ----
    You only need to call this method once, i.e, after the dataset has been downloaded
    and you have the `heartdisease.pkl` file, you don't need to call this method again.
    '''
    # Url to download the dataset from
    url = 'https://archive.ics.uci.edu/ml/machine-learning-databases/heart-disease/processed.cleveland.data'

    # Download the dataset
    print('Downloading processed.cleveland.data...')
    request.urlretrieve(url, 'processed.cleveland.data')
    print('Download complete.')

    # Parse data and save it as a pkl file.
    data_array = []
    # Open the file and put the values in a list.
    with open('processed.cleveland.data', 'r') as datafile:
        for line in datafile:
            try:
                data_array.append(list(map(float, line.split(','))))
            except ValueError:
                continue
    # Convert the list into a numpy array.
    heartdisease_data_array = np.array(data_array)
    # Save as a pkl file.
    pklhandler.save(heartdisease_data_array, 'heartdisease.pkl')

    # Delete unnecessary files.
    os.remove('processed.cleveland.data')
    print('Deleted unnecessary files.')

def load():
    '''
    Loads heart disease dataset from saved pickle file `heartdisease.pkl` to numpy arrays.
    Loads data without any preprocessing.

    Returns
    -------
    inputs : numpy.array
        297x13 numpy array. 297 training examples, each example having 13 inputs(columns).
        The 13 columns corresponds to: 
        :code:`age sex cp trestbps chol fbs restecg thalach exang oldpeak slope ca thal`.

        - age : Age in years
        - sex : 1=male, 0=female
        - cp : Chest pain type (1=typical-angina, 2=atypical-angina 3=non-anginal 4=asymptomatic)
        - trestbps :  Resting blood pressure in mmHg
        - chol : Serum cholesterol in mg/dl
        - fbs : Fasting blood sugar > 120 mg/dl? (1=true, 0=false)
        - restecg : Resting electrocardiographic results (0=normal, 1=ST-T-abnormality 2= left-ventricular-hypertrophy)
        - thalach : Maximum heart rate achieved 
        - exang : Exercise induced angina (1=yes, 0=no) 
        - oldpeak : ST depression induced by exercise relative to rest
        - slope: Slope of the peak exercise ST segment (1=upsloping 2=flat 3=downsloping)
        - ca : Number of major vessels colored by flourosopy (0-3)
        - thal: 3=normal, 6=fixed-defect, 7=reversable-defect

    outputs : numpy.array
        Numpy array with 297 elements. 
        
        - 0: < 50% diameter narrowing 
        - 1: > 50% diameter narrowing    

    Raises
    ------
        FileNotFoundError
            If `heartdisease.pkl` file does not exist, i.e, if the dataset was not 
            downloaded and saved using the :py:func:`~get` method.  
    '''
    # Load data from pkl file.
    heartdisease_data_array = pklhandler.load('heartdisease.pkl')   
    inputs =  heartdisease_data_array[:,:-1]
    outputs = (heartdisease_data_array[:,-1]>0)*1

    # return data
    return inputs, outputs




import os
import cProfile
from unittest.mock import patch
from functools import wraps

import matplotlib.pyplot as plt
from graphviz import Digraph 

import numpy as np


def _profile(test_func):
    '''
    Calls test function and profiles it.

    Parameters
    ----------
    test_func : function
        The function to test and profile.
    '''
    # Reset random seed
    np.random.seed(0)
    # Call the test function and profile it
    profiler = cProfile.Profile()
    profiler.runcall(test_func)
    profiler.dump_stats(test_func.__name__+'.dat') 


def pktest_graph(test_func):
    '''
    To test and profile function under pytest. Will prevent 
    :code:`matplotlib.pyplot.show()` from blocking other tests.

    Parameters
    ----------
    test_func : function
        The function to test and profile.
    '''
    # Create wrapper function for testing and profiling in pytest
    @wraps(test_func)
    def test_wrapper():
        # Close any open plots
        plt.close()
        plt.clf()

        with patch('matplotlib.pyplot.show') as show_func, patch('graphviz.Digraph.view') as view_func:
            # Run the test function
            _profile(test_func)
        
            # Test if graph worked
            if "PYTEST_CURRENT_TEST" in os.environ:
                assert show_func.called
                

    return test_wrapper


def pktest_nograph(test_func):
    '''
    To test and profile function under pytest.

    Parameters
    ----------
    test_func : function
        The function to test and profile.
    '''
    # Create wrapper function for testing and profiling in pytest
    @wraps(test_func)
    def test_wrapper():
        _profile(test_func)

    return test_wrapper


import numpy as np

def cross_validate(inputs, outputs, folds=5):
    '''
    Python generator function for making K-fold cross validation easier.

    Parameters
    ----------
    inputs : numpy.array
        Inputs/features of training data.
    outputs : numpy.array
        Outputs/targets of training data.

    Yields
    ------
    train_inputs : numpy.array
        Training data containing inputs.
    train_outputs : numpy.array
        Training data containing outputs.
    test_inputs : numpy.array
        Testing data containing inputs.
    test_outputs : numpy.array
        Testing data containing outputs.

    Example
    -------
    >>> import numpy as np
    >>> import pykitml as pk
    >>> 
    >>> # Mock training data
    ... x = np.arange(30).reshape((10, 3))
    >>> y = x + 10
    >>> 
    >>> # 5-fold cross validation
    ... # Training data is split into 5 blocks, each block takes its turn
    ... # to be the test data.
    ... for train_x, train_y, test_x, test_y in pk.cross_validate(x, y, 5):
    ...     print(train_x)
    ...     print(train_y)
    ...     print(test_x)
    ...     print(test_y)
    '''
    size = inputs.shape[0]
    block_size = size//folds
    remainder = size%folds

    # Calculate block sizes
    def get_block_size(block):
        if(block < remainder): return block_size+1
        else: return block_size

    block_sizes = [get_block_size(block) for block in range(folds)]

    # Calculate block indices
    block_indices = [sum(block_sizes[:block]) for block in range(folds)]

    # Generate blocks
    def make_block(i, array):
        start = block_indices[i]
        end = block_indices[i]+block_sizes[i]
        return array[start:end]

    for i in range(folds):
        # Create testing data
        test_inputs, test_outputs = make_block(i, inputs), make_block(i, outputs)

        # Create training data
        train_blocks_inputs = [make_block(j, inputs) for j in range(folds) if(j!=i)]
        train_inputs = np.concatenate(train_blocks_inputs, axis=0)
        train_blocks_outputs = [make_block(j, outputs) for j in range(folds) if(j!=i)]
        train_outputs = np.concatenate(train_blocks_outputs, axis=0)

        yield train_inputs, train_outputs, test_inputs, test_outputs



import setuptools

with open('README.md', 'r') as f:
    long_description = f.read()

setuptools.setup(
    name='pykitml',
    version='0.1.1',
    author='RainingComputers',
    author_email='vishnu.vish.shankar@gmail.com',
    description='Machine Learning library written in Python and NumPy.',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/RainingComputers/pykitml',
    packages=setuptools.find_packages(exclude=['docs', 'tests']),
    python_requires='>=3.5',
    install_requires=[
        'numpy', 'matplotlib', 'tqdm', 'graphviz'
    ],
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Development Status :: 3 - Alpha',
        'Topic :: Scientific/Engineering :: Artificial Intelligence'
    ],
    keywords='pykitml'
)


import os
from urllib import request

import numpy as np
from numpy import genfromtxt

from .. import pklhandler

'''
This module contains helper functions to download and load
the banknote dataset.
'''

def get():
    '''
    Downloads the banknote dataset from
    http://archive.ics.uci.edu/ml/datasets/banknote+authentication
    and saves it as a pkl file `banknote.pkl`.
    
    Raises
    ------
        urllib.error.URLError
            If internet connection is not available or the URL is not accessible.
        OSError
            If the file cannot be created due to a system-related error.
        KeyError
            If invalid/unknown type.

    Note
    ----
    You only need to call this method once, i.e, after the dataset has been downloaded
    and you have the `banknote.pkl` file, you don't need to call this method again.   
    '''
    # Url to download the dataset from
    url = 'http://archive.ics.uci.edu/ml/machine-learning-databases/00267/data_banknote_authentication.txt'
    
    # Download the dataset
    print('Downloading data_banknote_authentication.txt')
    request.urlretrieve(url, 'data_banknote_authentication.txt')
    print('Download complete.')

    # Parse the data and save it as a pkl file
    pklhandler.save(genfromtxt('data_banknote_authentication.txt', delimiter=','), 'banknote.pkl')

    # Delete unnecessary files
    os.remove('data_banknote_authentication.txt')
    print('Deleted unnecessary files.')

def load():
    '''
    Loads the banknote data from pkl file.
    
    The inputs have the following columns:

    - Variance of Wavelet Transformed image (continuous)
    - Skewness of Wavelet Transformed image (continuous) 
    - Curtosis of Wavelet Transformed image (continuous) 
    - Entropy of image (continuous) 

    The outputs are:

    - 0 = Real
    - 1 = Counterfeit

    Returns
    -------
    inputs_train : numpy.array
        1102x4 numpy array containing training inputs.
    outputs_train : numpy.array
        Numpy array of size 1102.
    inputs_test : numpy.array
        270x4 numpy array containing testing inputs.
    outputs_test : numpy.array
        Numpy array of size 270.
    
    '''
    data_array = pklhandler.load('banknote.pkl')

    # Separate data, positive and negative examples
    negative_examples = data_array[:762]
    positive_examples = data_array[762:]

    # Separate into training and testing
    negative_examples_test = negative_examples[:150]
    negative_examples_train = negative_examples[150:]
    positive_examples_test = positive_examples[:120]
    positive_examples_train = positive_examples[120:]

    # Join them to form training and testing dataset
    train = np.concatenate((negative_examples_train, positive_examples_train), axis=0)
    test = np.concatenate((negative_examples_test, positive_examples_test), axis=0)

    # Shuffle the dataset
    shuff_indices = np.arange(train.shape[0])
    np.random.shuffle(shuff_indices)    
    train = train[shuff_indices]
    shuff_indices = np.arange(test.shape[0])
    np.random.shuffle(shuff_indices)   
    test = test[shuff_indices]

    inputs_train = train[:, :-1]
    outputs_train = train[:, -1]
    inputs_test = test[:, :-1]
    outputs_test = test[:, -1]

    return inputs_train, outputs_train, inputs_test, outputs_test
        


import numpy as np

# ===============================================
# = Functions for Normalization/Feature-scaling =
# ===============================================

def get_minmax(array):
    '''
    Returns two row arrays, one array containing minimum values of each column
    and another one with maximum values.

    Parameters
    ----------
    array : numpy.array
        The array to get minimum and maximum values for.

    Returns
    -------
    array_min : numpy.array
        Array containing minimum values of each column.
    array_max : numpy.array
        Array containing maximum values of each column.
    '''
    return np.amin(array, axis=0), np.amax(array, axis=0)

def normalize_minmax(array, array_min, array_max, cols=[]):
    '''
    Normalizes columns of the array to between 0 and 1 using min-max
    normalization.

    Parameters
    ----------
    array : numpy.array
        The array to normalize.
    array_min : numpy.array
        Array containing minimum values of each column.
    array_max : numpy.array
        Array containing maximum values of each column.
    cols : list
        The columns to normalize. If the list is empty (default),
        all columns will be normalized.

    Returns
    -------
    numpy.array
        The normalized array.

    Note
    ----
    You can use :py:func:`~get_minmax` function to get :code:`array_min`
    and :code:`array_max` parameters.
    '''
    normalized_array = array.astype(float)
    all_normalized = (array - array_min) / (array_max - array_min)

    if(len(cols) == 0):
        # Normalize all columns
        normalized_array = all_normalized
    elif(array.ndim == 1):
        # Normalize only specified columns, 1D array
        normalized_array[cols] = all_normalized[cols]
    else:
        # Normalize onlt specified columns, 2D array
        normalized_array[:, cols] = all_normalized[:, cols]

    return normalized_array

def denormalize_minmax(array, array_min, array_max, cols=[]):
    '''
    Denormalizes columns of a min-max normalized array.

    Parameters
    ----------
    array : numpy.array
        The array to denormalize.
    array_min : numpy.array
        Array containing minimum values of each column.
    array_max : numpy.array
        Array containing maximum values of each column.
    cols : list
        The columns to normalize. If the list is empty (default),
        all columns will be denormalized.

    Returns
    -------
    numpy.array
        The denormalized array.

    Note
    ----
    You can use :py:func:`~get_minmax` function to get :code:`array_min`
    and :code:`array_max` parameters.
    '''
    denormalized_array = array.astype(float)
    all_denormalized = (array * (array_max - array_min)) + array_min

    if(len(cols) == 0):
        # Deormalize all columns
        denormalized_array = all_denormalized
    elif(array.ndim == 1):
        # Denormalize only specified columns, 1D array
        denormalized_array[cols] = all_denormalized[cols]
    else:
        # Denormalize onlt specified columns, 2D array
        denormalized_array[:, cols] = all_denormalized[:, cols]

    return denormalized_array


def get_meanstd(array):
    '''
    Returns two row arrays, one array containing mean of each column
    and another one with standard deviation of each column.

    Parameters
    ----------
    array : numpy.array
        The array to get mean and standard deviation values for.

    Returns
    -------
    array_mean : numpy.array
        Array containing mean values of each column.
    array_stddev : numpy.array
        Array containing standard deviation values of each column.
    '''
    return np.mean(array, axis=0), np.std(array, axis=0)

def normalize_mean(array, array_mean, array_stddev, cols=[]):
    '''
    Normalizes columns of the array with mean normalization.

    Parameters
    ----------
    array : numpy.array
        The array to normalize.
    array_mean : numpy.array
        Array containing mean values of each column.
    array_stddev : numpy.array
        Array containing standard deviation values of each column.
    cols : list
        The columns to normalize. If the list is empty (default),
        all columns will be normalized.


    Returns
    -------
    numpy.array
        The normalized array.

    Note
    ----
    You can use :py:func:`~get_meanstd` function to get :code:`array_mean`
    and :code:`array_stddev` parameters.
    '''
    normalized_array = array.astype(float)
    all_normalized = (array-array_mean)/array_stddev

    if(len(cols) == 0):
        # Normalize all columns
        normalized_array = all_normalized
    elif(array.ndim == 1):
        # Normalize only specified columns, 1D array
        normalized_array[cols] = all_normalized[cols]
    else:
        # Normalize only specified columns, 2D array
        normalized_array[:, cols] = all_normalized[:, cols]

    return normalized_array

def denormalize_mean(array, array_mean, array_stddev, cols=[]):
    '''
    Denormalizes a mean normalized array.

    Parameters
    ----------
    array : numpy.array
        The array to denormalize.
    array_mean : numpy.array
        Array containing mean values of each column.
    array_stddev : numpy.array
        Array containing standard deviation values of each column.

    Returns
    -------
    numpy.array
        The denormalized array.

    Note
    ----
    You can use :py:func:`~get_meanstd` function to get :code:`array_mean`
    and :code:`array_stddev` parameters.
    '''
    denormalized_array = array.astype(float)
    all_denormalized = (array*array_stddev) + array_mean

    if(len(cols) == 0):
        # Denormalize all columns
        denormalized_array = all_denormalized
    elif(array.ndim == 1):
        # Denormalize only specified columns, 1D array
        denormalized_array[cols] = all_denormalized[cols]
    else:
        # Denormalize onlt specified columns, 2D array
        denormalized_array[:, cols] = all_denormalized[:, cols]

    return denormalized_array


import os
from urllib import request

import numpy as np

from .. import pklhandler

'''
This module contains helper functions to download the sonar dataset.
'''

def get():
    '''
    Downloads sonar dataset from
    https://archive.ics.uci.edu/ml/datasets/Connectionist+Bench+(Sonar,+Mines+vs.+Rocks)
    and saves it as a pkl file `sonar.pkl`.

    Raises
    ------
        urllib.error.URLError
            If internet connection is not available or the URL is not accessible.
        OSError
            If the files cannot be created due to a system-related error.
        KeyError
            If invalid/unknown type.

    Note
    ----
    You only need to call this method once, i.e, after the dataset has been downloaded
    and you have the `sonar.pkl` file, you don't need to call 
    this method again.
    '''
    # Url to download the dataset from
    url = 'https://archive.ics.uci.edu/ml/machine-learning-databases/undocumented/connectionist-bench/sonar/sonar.all-data'

    # Download the dataset
    print('Downloading sonar.all-data...')
    request.urlretrieve(url, 'sonar.all-data')
    print('Download complete.')

    out_dict = {
        'R\n':0, 'M\n':1
    }

    # Parse data and save it as pkl file
    data_array = []
    # Open the file and put the values in a list.
    with open('sonar.all-data', 'r') as datafile:
        for line in datafile:
            values = line.split(',')
            values[-1] = out_dict[values[-1]]
            data_array.append(list(map(float, values)))
    # Convert the list to numpy array
    sonar_data_array = np.array(data_array)
    # Save it as a pkl file
    pklhandler.save(sonar_data_array, 'sonar.pkl')

    # Delete files
    os.remove('sonar.all-data')

def load():
    '''
    Loads the adult dataset from `sonar.pkl` file.

    Each pattern is a set of 60 numbers in the range 0.0 to 1.0. 
    Each number represents the energy within a particular frequency band, 
    integrated over a certain period of time. The integration aperture for 
    higher frequencies occur later in time, since these frequencies are 
    transmitted later during the chirp.

    The label associated with each record contains the letter 
    "R" if the object is a rock and "M" if it is a mine (metal cylinder).

    Returns
    -------
    inputs_train : numpy.array
        190x60 numpy array containing training inputs.
    outputs_train : numpy.array
        Numpy array of size 190.
    inputs_test : numpy.array
        18x60 numpy array containing testing inputs.
    outputs_test : numpy.array
        Numpy array of size 18.

    Raises
    ------
        filesNotFoundError
            If `sonar.pkl` file does not exist, 
            i.e, if the dataset was not downloaded and saved using the 
            :py:func:`~get` method.     

    '''
    # Load the data from pkl file
    sonar_data_array = pklhandler.load('sonar.pkl')

    # Split into train and test
    train_neg = sonar_data_array[0:90]
    train_pos = sonar_data_array[97:197]
    test_neg = sonar_data_array[90:97]
    test_pos = sonar_data_array[197:208]

    # Shuffle the dataset, join neg and pos examples
    train = np.concatenate((train_pos, train_neg), axis=0)
    np.random.shuffle(train)
    test = np.concatenate((test_pos, test_neg), axis=0)
    np.random.shuffle(test)

    # Split the dataset into inputs and outputs
    inputs_train = train[:, :-1]
    outputs_train = train[:, -1]
    inputs_test = test[:, :-1]
    outputs_test = test[:, -1]

    # return
    return inputs_train, outputs_train, inputs_test, outputs_test



import numpy as np

from ._minimize_model import MinimizeModel
from ._classifier import Classifier
from ._regressor import Regressor
from . import _functions

class NeuralNetwork(MinimizeModel, Classifier, Regressor):
    '''
    This class implements Feed-forward Neural Network.
    '''

    def __init__(self, layer_sizes, reg_param=0, config='leakyrelu-softmax-cross_entropy'):
        '''
        Parameters
        ----------
        layer_sizes : list
            A list of integers describing the number of layers and the number of neurons in each
            layer. For e.g. :code:`[784, 100, 100, 10]` describes a neural network with one input
            layer having 784 neurons, two hidden layers having 100 neurons each and a output layer
            with 10 neurons.
        reg_param : int
            Regularization parameter for the network, also known as 'weight decay'.
        config : str
            The config string describes what activation functions and cost function to use for
            the network. The string should contain three function names seperated with '-' 
            character and should follow the order:
            :code:`'<hidden_layer_activation_func>-<output_layer_activation_func>-<cost_function>'`.
            For e.g. :code:`'relu-softmax-cross_entropy'` tells the class to use `relu` as the
            activation function for input and hidden layers, `softmax` for output layer and
            `cross entropy` for the cost function.

            List of available activation functions:
            :code:`leakyrelu`, :code:`relu`, :code:`softmax`, :code:`tanh`, :code:`sigmoid`, :code:`identity`.

            List of available cost functions:
            :code:`mse` (Mean Squared Error), :code:`cross_entropy` (Cross Entropy), :code:`huber` (Huber loss).

        Raises
        ------
        AttributeError
            If invalid config string.
        '''
        # Get function names from the config string, the config string contains
        # list of functions to be used in the neural network, seperated with '-' character.
        # syntax: 'hidden_layers-output_layer-cost function', eg: 'relu-sigmoid-cross_entropy', 
        func_names = config.split('-')
        func_names_prime = [func_name + '_prime' for func_name in func_names]
        
        # Initialize regularization parameter
        self._reg_param = reg_param
        self._reg_param_half = reg_param/2

        # Initialize functions
        self._activ_func = getattr(_functions, func_names[0])
        self._activ_func_prime = getattr(_functions, func_names_prime[0])
        self._output_activ_func = getattr(_functions, func_names[1])
        self._output_activ_func_prime = getattr(_functions, func_names_prime[1])
        self._cost_func = getattr(_functions, func_names[2])
        self._cost_func_prime = getattr(_functions, func_names_prime[2])

        # Store activation function names
        self._functs = func_names

        # Store layer sizes
        self._lsize = layer_sizes

        # Number of layers
        self._nlayers = len(layer_sizes)

        # Intialize parameters
        weights = [np.array([])] * self.nlayers
        biases = [np.array([])] * self.nlayers
        # Loop through each layer and initialize random weights and biases
        for l in range(1, self.nlayers):
            layer_size = layer_sizes[l]
            input_layer_size = layer_sizes[l-1]
            epsilon = np.sqrt(6 / (layer_size + input_layer_size))
            weights[l] = np.random.uniform(-epsilon, epsilon, (layer_size, input_layer_size))
            biases[l] = np.random.uniform(-epsilon, epsilon, (layer_size))
        # Put parameters in numpy dtype=object array
        self._params = np.array(
            [np.array(weights, dtype=object), np.array(biases, dtype=object)],
            dtype=object
        )  

        # List of numpy arrays for storing temporary values
        self.z = [np.array([])] * self.nlayers
        self.a = [np.array([])] * self.nlayers

    def __repr__(self):
        return 'Layers: '+str(self._lsize)+' Config: '+str(self._functs)

    @property
    def _mparams(self):
        return self._params
    
    @_mparams.setter
    def _mparams(self, mparams):
        self._params = mparams

    @property
    def _cost_function(self):
        return self._cost_func

    @property
    def _out_size(self):
        return self._lsize[-1]

    @property
    def nlayers(self):
        '''
        The number of layers in the network.
        '''      
        return self._nlayers

    def feed(self, input_data):
        # Constants
        W = 0 # Weights
        B = 1 # Biases

        # Set inputs
        self.a[0] = input_data

        # Feed through hidden layers
        for l in range(1, self.nlayers-1):
            self.z[l] = (self.a[l-1]@self._params[W][l].T) + self._params[B][l]
            self.a[l] = self._activ_func(self.z[l])

        # Feed thorugh output layer
        self.z[-1] = (self.a[-2]@self._params[W][-1].T) + self._params[B][-1]
        self.a[-1] = self._output_activ_func(self.z[-1])

    def get_output(self):
        return self.a[-1].squeeze()

    def _backpropagate(self, index, target):
        # Constants
        W = 0 # Weights        
        
        # Used to hold activation_function'(z[l]) where z[l] = w[l]*a[l-1] + b[l] 
        da_dz = [np.array([])] * self.nlayers

        # Used to hold partial derivatives of cost function w.r.t parameters
        dc_dw = [np.array([])] * self.nlayers
        dc_db = [np.array([])] * self.nlayers
        
        # Calculate activation_function'(z)
        def calc_da_dz(l):
            da_dz[l] = self._activ_func_prime(self.z[l][index], self.a[l][index])
        
        # Calculate the partial derivatives of the cost w.r.t all the biases of layer 
        # 'l' (NOT for output layer)
        def calc_dc_db(l):
            dc_db[l] = (dc_db[l+1] @ self._params[W][l+1]) * da_dz[l]

        # Calculate the partial derivatives of the cost w.r.t all the weights of layer 'l'
        def calc_dc_dw(l):
            dc_dw[l] = np.multiply.outer(dc_db[l], self.a[l-1][index])
            # Regularization
            dc_dw[l] += self._reg_param*self._params[W][l]

        # Calculate the partial derivatives of the cost function w.r.t the output layer's 
        # activations, weights, biases
        da_dz[-1] = self._output_activ_func_prime(self.z[-1][index], self.a[-1][index])
        dc_db[-1] = self._cost_func_prime(self.a[-1][index], target) * da_dz[-1]
        calc_dc_dw(-1)

        # Calculate the partial derivatives of the cost function w.r.t the hidden layers'
        # activations, weights, biases
        for l in range(self.nlayers-2, 0, -1):
            calc_da_dz(l)
            calc_dc_db(l)
            calc_dc_dw(l)

        # return gradient
        return np.array(
            [np.array(dc_dw, dtype=object), np.array(dc_db, dtype=object)],
            dtype=object
        )

    @property
    def bptt(self):
        return False

    def _get_norm_weights(self):
        # If regularization is zero
        if(self._reg_param == 0): return 0
        
        # else
        W = 0
        norm = 0
        # Calculate norm of the weights
        for l in range(self.nlayers):
            norm += (self._reg_param_half)*(self._params[W][l]**2).sum()
        
        return norm
